__all__ = ()

from json import JSONDecodeError
from os import get_terminal_size
from reprlib import repr as short_repr

from scarletio import from_json
from scarletio.web_common import FormData
from scarletio.web_common.form_data import FORM_DATA_FIELD_TYPE_JSON


TYPE_NAME_UNEXPECTED = 'Object'
TYPE_NAME_STRING = 'String'
TYPE_NAME_BINARY = 'Binary'
TYPE_NAME_BOOLEAN = 'Boolean'
TYPE_NAME_INTEGER = 'Integer'
TYPE_NAME_FLOAT = 'Float'
TYPE_NAME_LIST = 'List'
TYPE_NAME_HASH_MAP = 'HashMap'
TYPE_NAME_FORM_DATA = 'FormData'

MODIFIER_LENGTH = 'length'

VALUE_INDENT = '    '
VALUE_NONE = 'null'
VALUE_BOOLEAN_TRUE = 'true'
VALUE_BOOLEAN_FALSE = 'false'
VALUE_QUOTE = '"'
VALUE_PAYLOAD = 'payload'

ESCAPABLE = {VALUE_QUOTE, '\\', '\r', '\n', '\t'}

LINE_WIDTH_DEFAULT = 120

STRING_MIN_LINE_LENGTH = 40
STRING_BREAK_TO_MULTI_LINE_OVER = 60
STRING_MAX_RENDER_LENGTH = 6000


def reconstruct_payload(payload):
    """
    Tries to reconstruct the given payload.
    
    Parameters
    ----------
    payload : `object`
        The payload to try to reconstruct.
    
    Returns
    -------
    reconstructed_value : `None | str`
    """
    if payload is None: # nothing to do
        return None
    
    line_width = _get_terminal_line_width()
    
    into = [VALUE_PAYLOAD, ' = ']
    used_characters = len(VALUE_PAYLOAD) + 3
    
    if isinstance(payload, str):
        into = reconstruct_json_into(payload, into, 0, line_width, used_characters)
    
    elif isinstance(payload, FormData):
        into = reconstruct_form_data_into(payload, into, line_width)
    
    elif isinstance(payload, bytes):
        into = reconstruct_binary_into(payload, into, 0, line_width, used_characters)
    
    else:
        into = reconstruct_unexpected_into(payload, into, 0, line_width, len(VALUE_PAYLOAD) + 3)
    
    return ''.join(into)


def reconstruct_json_into(value, into, indent, line_width, used_characters):
    """
    Reconstructs a json payload extending the given `into` list.
    
    Parameters
    ----------
    value : `str`
        Json payload data to reconstruct.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        Additionally used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    try:
        json_data = from_json(value)
    except JSONDecodeError:
        into = reconstruct_string_into(value, into, indent, False, line_width, used_characters)
    else:
        into = reconstruct_value_into(json_data, into, indent, False, line_width, used_characters)
    
    return into


def reconstruct_value_into(value, into, indent, value_is_file, line_width, used_characters):
    """
    Reconstructs a value extending the given `into` list.
    
    Parameters
    ----------
    value : `None`, `str`, `list`, `dict`, `int`, `float`, `bool`
        The deserialized json value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    value_is_file : `bool`
        Whether the value is a file and should not shown.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        Additionally used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    if value is None:
        into = reconstruct_null_into(into, indent, line_width, used_characters)
    
    elif isinstance(value, bool):
        into = reconstruct_boolean_into(value, into, indent, line_width, used_characters)

    elif isinstance(value, str):
        into = reconstruct_string_into(value, into, indent, value_is_file, line_width, used_characters)
    
    elif isinstance(value, int):
        into = reconstruct_integer_into(value, into, indent, line_width, used_characters)
    
    elif isinstance(value, float):
        into = reconstruct_float_into(value, into, indent, line_width, used_characters)
    
    elif isinstance(value, list):
        into = reconstruct_list_into(value, into, indent, line_width)
    
    elif isinstance(value, dict):
        into = reconstruct_hash_map_into(value, into, indent, line_width)
    
    elif isinstance(value, bytes):
        # bytes is used at form data
        into = reconstruct_binary_into(value, into, indent, line_width, used_characters)
    
    elif isinstance(value, set):
        # Set can be a substitution to list.
        into = reconstruct_list_into([*value], into, indent, line_width)
    
    else:
        into = reconstruct_unexpected_into(value, into, indent, line_width, used_characters)
    
    return into


def reconstruct_null_into(into, indent, line_width, used_characters):
    """
    Reconstructs a null value to the given `into` list.
    
    Parameters
    ----------
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        The amount of used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    should_wrap = line_width < (
        used_characters +
        indent * len(VALUE_INDENT) +
        len(VALUE_NONE)
    )
    
    if should_wrap:
        into = _wrap_begin(into, indent)
    
    into.append(VALUE_NONE)
    
    if should_wrap:
        into = _wrap_end(into, indent)
    
    return into


def reconstruct_boolean_into(value, into, indent, line_width, used_characters):
    """
    Reconstructs a boolean value to the given `into` list.
    
    Parameters
    ----------
    value : `bool`
        The boolean value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        The amount of used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    if value:
        value_representation = VALUE_BOOLEAN_TRUE
    else:
        value_representation = VALUE_BOOLEAN_FALSE
    
    should_wrap = line_width < (
        used_characters +
        indent * len(VALUE_INDENT) +
        len(value_representation)
    )
    
    if should_wrap:
        into = _wrap_begin(into, indent)
    
    into.append(value_representation)
    
    if should_wrap:
        into = _wrap_end(into, indent)
    
    return into


def _reconstruct_string_as_single_line_into(value, into):
    """
    Reconstructs a string as single line extending the given `into` list.
    
    Parameters
    ----------
    value : `str`
        The string value.
    into : `list<str>`
        A list to extend it's content.
    
    Returns
    -------
    into : `list<str>`
    """
    into.append(_get_string_representation(value))
    return into


def _reconstruct_string_as_multi_line_into(value, into, indent, line_width):
    """
    Reconstructs a string as multi line extending the given `into` list.
    
    Parameters
    ----------
    value : `str`
        The string value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    
    Returns
    -------
    into : `list<str>`
    """
    into.append('(\n')
    
    string_indent = indent + 1
    
    length_limit = line_width - string_indent * len(VALUE_INDENT)
    if length_limit < STRING_MIN_LINE_LENGTH:
        length_limit = STRING_MIN_LINE_LENGTH
    
    for representation in _iter_string_representation_range(value, length_limit):
        for counter in range(string_indent):
            into.append(VALUE_INDENT)
        
        into.append(representation)
        into.append('\n')
    
    for counter in range(indent):
        into.append(VALUE_INDENT)
    
    into.append(')')
    
    return into


def _reconstruct_string_as_annotation_into(value, into, indent, line_width, used_characters):
    """
    Reconstructs a string as an annotation extending the given `into` list.
    
    Parameters
    ----------
    value : `str`
        The string value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        The amount of used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    length_representation = str(len(value))
    estimated_length = (
        used_characters +
        indent * len(VALUE_INDENT) +
        len(TYPE_NAME_STRING) + 5 + len(MODIFIER_LENGTH) + len(length_representation)
    )
    
    should_wrap = line_width < estimated_length
    
    if should_wrap:
        into = _wrap_begin(into, indent)
    
    into.append(TYPE_NAME_STRING)
    into.append('<')
    into.append(MODIFIER_LENGTH)
    into.append(' = ')
    into.append(str(len(value)))
    into.append('>')
    
    if should_wrap:
        into = _wrap_end(into, indent)
    
    return into


def reconstruct_string(value, indent, value_is_file, line_width, used_characters):
    """
    Reconstructs the given string.
    
    Parameters
    ----------
    value : `str`
        The string value.
    indent : `int`
        The amount of indents to add.
    value_is_file : `bool`
        Whether the value is a file and should not be shown.
    used_characters : `int`
        The amount of used characters.
    
    Returns
    -------
    result : `str`
    """
    return ''.join(reconstruct_string_into(value, [], indent, value_is_file, line_width, used_characters))


def reconstruct_string_into(value, into, indent, value_is_file, line_width, used_characters):
    """
    Reconstructs a string value to the given `into` list.
    
    Parameters
    ----------
    value : `str`
        The string value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    value_is_file : `bool`
        Whether the value is a file and should not be shown.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        The amount of used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    if value_is_file or (len(value) > STRING_MAX_RENDER_LENGTH):
        into = _reconstruct_string_as_annotation_into(value, into, indent, line_width, used_characters)
    
    else:
        if _is_string_length_over(value, line_width - used_characters - len(VALUE_INDENT) * indent):
            into = _reconstruct_string_as_multi_line_into(value, into, indent, line_width)
        
        else:
            into = _reconstruct_string_as_single_line_into(value, into)
    
    return into


def reconstruct_integer_into(value, into, indent, line_width, used_characters):
    """
    Reconstructs an integer value to the given `into` list.
    
    Parameters
    ----------
    value : `int`
        The integer value.
    into : `list<str>`
        A list to extend it's content.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    
    Returns
    -------
    into : `list<str>`
    """
    representation = repr(value)
    
    should_wrap = line_width < (
        used_characters +
        indent * len(VALUE_INDENT) +
        len(representation)
    )
    
    if should_wrap:
        into = _wrap_begin(into, indent)
    
    into.append(representation)
    
    if should_wrap:
        into = _wrap_end(into, indent)
    
    return into


def reconstruct_float_into(value, into, indent, line_width, used_characters):
    """
    Reconstructs a float value to the given `into` list.
    
    Parameters
    ----------
    value : `float`
        The float value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        The amount of used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    value_representation = repr(value)
    
    should_wrap = line_width < (
        used_characters +
        indent * len(VALUE_INDENT) +
        len(value_representation)
    )
    
    if should_wrap:
        into = _wrap_begin(into, indent)
    
    into.append(value_representation)
    
    if should_wrap:
        into = _wrap_end(into, indent)
    
    return into


def reconstruct_list_into(value, into, indent, line_width):
    """
    Reconstructs a list value extending the given `into` list.
    
    Parameters
    ----------
    value : `list`
        The list value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    
    Returns
    -------
    into : `list<str>`
    """
    length = len(value)
    
    into.append('[')
    
    if length:
        into.append('\n')
        
        element_indent = indent + 1
        
        for index, list_element in enumerate(value):
            for counter in range(element_indent):
                into.append(VALUE_INDENT)
            
            index_string = str(index)
            into.append(index_string)
            into.append(': ')
            
            reconstruct_value_into(list_element, into, element_indent, False, line_width, len(index_string) + 3)
            
            into.append(',\n')
        
        for counter in range(indent):
            into.append(VALUE_INDENT)
    
    into.append(']')
    return into


def reconstruct_hash_map_into(value, into, indent, line_width):
    """
    Reconstructs a hash map value extending the given `into` list.
    
    Parameters
    ----------
    value : `dict`
        The hash map value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    
    Returns
    -------
    into : `list<str>`
    """
    length = len(value)
    
    into.append('{')
    
    if length:
        into.append('\n')
        item_indent = indent + 1
        
        for item_key, item_value in sorted(value.items(), key = _hash_map_key_sort_key):
            for counter in range(item_indent):
                into.append(VALUE_INDENT)
            
            item_key_representation = reconstruct_string(item_key, item_indent, False, line_width, 3)
            into.append(item_key_representation)
            into.append(': ')
            
            if item_key_representation.endswith(')'):
                used_characters = 1 + 3
            else:
                used_characters = len(item_key_representation) + 3
            
            reconstruct_value_into(item_value, into, item_indent, False, line_width, used_characters)
            
            into.append(',\n')
        
        for counter in range(indent):
            into.append(VALUE_INDENT)
    
    into.append('}')
    
    return into


def reconstruct_unexpected_into(value, into, indent, line_width, used_characters):
    """
    Reconstructs a value with an unexpected type to the given `into` list.
    
    Parameters
    ----------
    value : `object`
        Le value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        The amount of used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    value_representation = short_repr(value)
    
    should_wrap = line_width < (
        used_characters +
        indent * len(VALUE_INDENT) +
        len(TYPE_NAME_UNEXPECTED) + 2 + len(value_representation)
    )
    
    if should_wrap:
        into = _wrap_begin(into, indent)
    
    into.append(TYPE_NAME_UNEXPECTED)
    into.append(': ')
    into.append(value_representation)
    
    if should_wrap:
        into = _wrap_end(into, indent)
    
    return into


def reconstruct_binary_into(value, into, indent, line_width, used_characters):
    """
    Reconstructs a binary value to the given `into` list.
    
    Parameters
    ----------
    value : `binary`
        The binary value.
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    line_width : `int`
        Allowed line width.
    used_characters : `int`
        The amount of used characters.
    
    Returns
    -------
    into : `list<str>`
    """
    length_representation = str(len(value))
    
    should_wrap = line_width < (
        used_characters +
        indent * len(VALUE_INDENT) +
        len(TYPE_NAME_BINARY) + 5 + len(MODIFIER_LENGTH) + len(length_representation)
    )
    
    if should_wrap:
        into = _wrap_begin(into, indent)
    
    into.append(TYPE_NAME_BINARY)
    into.append('<')
    into.append(MODIFIER_LENGTH)
    into.append(' = ')
    into.append(str(len(value)))
    into.append('>')
    
    if should_wrap:
        into = _wrap_end(into, indent)
    
    return into


def reconstruct_form_data_into(value, into, line_width):
    """
    Tries to reconstruct form data into the given `into` list.
    
    Parameters
    ----------
    value : ``FormData``
        The form data to reconstruct.
    into : `list<str>`
        A list to extend it's content.
    line_width : `int`
        Allowed line width.
    
    Returns
    -------
    into : `list<str>`
    """
    into.append(TYPE_NAME_FORM_DATA)
    
    fields = value.fields
    
    into.append('(')
    
    if fields:
        into.append('{\n')
        
        for index, field in enumerate(fields):
            field_type_options = field.type_options
            
            field_name = field_type_options['name']
            file_name = field_type_options.get('file_name', None)
            
            into.append(VALUE_INDENT)
            index_representation = str(index)
            into.append(index_representation)
            into.append(': ')
            
            #                        index              + ': ' + {field_name} + (',' | '(')
            used_characters = len(index_representation) +  2   +                     1
            
            field_name_representation = reconstruct_string(field_name, 1, False, line_width, used_characters)
            into.append(field_name_representation)
            
            if (file_name is not None) and (file_name != field_name):
                
                if field_name_representation.endswith(')'):
                    #                 ')' + ' | ' + {file_name} + ': ' + (',' | '(')
                    used_characters =  1  +   3   +                2   +      1
                else:
                    #                          index           + ': ' +         field_name             + ' | ' + {file_name} + ': ' + (',' | '(')
                    used_characters = len(index_representation) + 2   + len(field_name_representation) +   3   +                2   +      1
                
                into.append(' | ')
                file_name_representation = reconstruct_string(file_name, 1, False, line_width, used_characters)
                into.append(file_name_representation)
            else:
                file_name_representation = None
            
            into.append(': ')
            
            if (file_name_representation is not None) and file_name_representation.endswith(')'):
                #                 ")" + ': ' + (',' | '(')
                used_characters =  1  +  2   +      1
            else:
                if (file_name_representation is None):
                    used_characters = 0
                else:
                    #                 ' | ' +          field_name
                    used_characters =   3   + len(file_name_representation)
                
                if field_name_representation.endswith(')'):
                    #                 ')' +  ?(' | ' + field_name) + ': ' + (',' | '(') + {field_value}
                    used_characters =  1  +      used_characters   +  3   +      1
                else:
                    #                           index           + ': ' +           field_name           + ?(' | ' + field_name) + ': ' + {field_value} + (',' | '(')
                    used_characters = len(index_representation) +  2   + len(field_name_representation) +     used_characters   +  2   +                      1
            
            reconstruct_value_into(
                field.value, into, 1, field.type != FORM_DATA_FIELD_TYPE_JSON, line_width, used_characters
            )
            into.append(',\n')
        
        into.append('}')
    
    into.append(')')
    
    return into


def _hash_map_key_sort_key(item):
    """
    Used to sort hash maps based on their key.
    
    Parameters
    ----------
    item : `tuple` (`str`, `object`)
        An item of a hash map.
    
    Returns
    -------
    key : `str`
    """
    return item[0]


def _get_terminal_line_width():
    """
    Gets the terminal's line's width.
    
    Returns
    -------
    line_width : `int`
    """
    try:
        terminal_size = get_terminal_size()
    except OSError:
        line_width = LINE_WIDTH_DEFAULT
    else:
        line_width = terminal_size.columns
    
    return line_width


def _get_string_representation(string):
    """
    Gets the string's representation.
    
    Parameters
    ----------
    string : `str`
        Teh string to represent.
    
    Returns
    -------
    representation : `str`
    """
    return ''.join([
        VALUE_QUOTE,
        *(_get_character_representation(character) for character in string),
        VALUE_QUOTE,
    ])


def _iter_string_representation_range(string, length_limit):
    """
    Yields `length_limit` representation of the given `string`.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    string : `str`
        String to represent.
    length_limit : `int`
        Maximal allowed length. Should be at least `12`.
    
    Yields
    ------
    representation : `str`
    """
    start = 0
    length = len(string)
    
    while start < length:
        representation, start = _get_string_representation_range(string, start, length_limit)
        yield representation


def _get_string_representation_range(string, start, length_limit):
    """
    Gets string representation fulfilling the given range.
    
    Parameters
    ----------
    string : `str`
        The string to represent.
    start : `int`
        First character to represent.
    length_limit : `int`
        Maximal allowed length. Should be at least `12`.
    
    Returns
    -------
    representation : `str`
    end : `int`
    """
    result = [VALUE_QUOTE]
    length = len(VALUE_QUOTE) << 1
    
    for index in range(start, len(string)):
        character = string[index]
        length += _get_character_representation_length(character)
        # If `length_limit` is low make sure we add at least 1 character
        if (length > length_limit) and (start != index):
            end = index
            break
        
        result.append(_get_character_representation(character))
        
    else:
        end = len(string)
    
    result.append(VALUE_QUOTE)
    return ''.join(result), end


def _is_string_length_over(string, length_limit):
    """
    Gets the length of the given string if it would be quoted.
    
    Parameters
    ----------
    string : `str`
        The string to check.
    length_limit : `int`
        Max allowed length.
    
    Returns
    -------
    is_over : `bool`
    """
    length = len(VALUE_QUOTE) << 1
    
    for character in string:
        length += _get_character_representation_length(character)
        if length > length_limit:
            return True
    
    return False


def _get_character_representation_escaped_length(character):
    """
    Gets the character's representation length if it would be escaped.
    
    Parameters
    ----------
    character : `str`
        Character to get its escaped length of.
    
    Returns
    -------
    length : `int`
    """
    character_int = ord(character)
    if character_int >= 0x10000 or character_int < 0:
        length = 2 + 8
    elif character_int >= 0x100:
        length = 2 + 4
    else:
        length = 2 + 2
    return length


def _get_character_representation_length(character):
    """
    Returns the character's length.
    
    Parameters
    ----------
    character : `str`
        Character to get its length of.
    
    Returns
    -------
    length : `int`
    """
    if character in ESCAPABLE:
        length = 2
    elif character.isprintable():
        length = 1
    else:
        length = _get_character_representation_escaped_length(character)
    
    return length


def _get_character_representation_escaped(character):
    """
    Escapes the given character.
    
    Parameters
    ----------
    character : `str`
        The character to escape.
    
    Returns
    -------
    result : `str`
    """
    character_int = ord(character)
    if character_int >= 0x10000 or character_int < 0:
        prefix = '\\U'
        hexadecimal_length = 8
    
    elif character_int >= 0x100:
        prefix = '\\u'
        hexadecimal_length = 4
    
    else:
        prefix = '\\x'
        hexadecimal_length = 2
    
    result = [prefix]
    for index in reversed(range(hexadecimal_length)):
        result.append('0123456789abcdef'[(character_int >> (index << 2)) & 0x0f])
    
    return ''.join(result)


def _get_character_representation(character):
    """
    Gets the character's representation.
    
    Parameters
    ----------
    character : `str`
        Character to represent.
    
    Returns
    --------
    representation : `str`
    """
    if character == VALUE_QUOTE:
        representation = '\\' + VALUE_QUOTE
    
    elif character == '\n':
        representation = '\\n'
    elif character == '\r':
        representation = '\\r'
    elif character == '\t':
        representation = '\\t'
    elif character == '\\':
        representation = '\\\\'
    elif character.isprintable():
        representation = character
    else:
        representation = _get_character_representation_escaped(character)
    
    return representation


def _wrap_begin(into, indent):
    """
    Begins wrapping.
    
    Parameters
    ----------
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    
    Returns
    -------
    into : `list<tr>`
    """
    into.append('(\n')
        
    for index in range(indent + 1):
        into.append(VALUE_INDENT)
    
    return into


def _wrap_end(into, indent):
    """
    Ends wrapping.
    
    Parameters
    ----------
    into : `list<str>`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    
    Returns
    -------
    into : `list<tr>`
    """
    into.append('\n')
    
    for index in range(indent):
        into.append(VALUE_INDENT)
    
    into.append(')')
    
    return into



__all__ = ('StringSelectOption',)

import reprlib

from scarletio import RichAttributeErrorBaseType

from ..shared_fields import parse_emoji, put_emoji_into, validate_emoji

from .fields import (
    parse_default, parse_description, parse_label, parse_value, put_default, put_description, put_label,
    put_value, validate_default, validate_description, validate_label, validate_value
)


class StringSelectOption(RichAttributeErrorBaseType):
    """
    An option of a string select component.
    
    Attributes
    ----------
    default : `bool`
        Whether this option is the default one.
    
    description : `None`, `str`
        Description of the option.
    
    emoji : `None`, ``Emoji``
        Emoji on the option if applicable.
    
    label : `str`
        Label of the option.
    
    value : `str`
        Identifier value of the option.
    """
    def __new__(cls, value, label = ..., emoji = ..., *, default = ..., description = ...):
        """
        Creates a new string select option with the given parameters.
        
        Parameters
        ----------
        value : `str`
            The option's value.
        
        label : `None | str`, Optional
            Label of the component option.
            
            Defaults to the `value` parameter if not given or if given as `None`.
        
        emoji : `None`, ``Emoji``, Optional
            Emoji of the option if applicable.
        
        default : `bool`, Optional (Keyword only)
            Whether this the the default option. Defaults to `False`.
        
        description : `None`, `str`, Optional (Keyword only)
            Description of the component option.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        value = validate_value(value)
        
        # default
        if default is ...:
            default = False
        else:
            default = validate_default(default)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # label
        if label is ...:
            label = None
        else:
            label = validate_label(label)
        
        # Post validation
        if label is None:
            label = value
        
        # Construct
        self = object.__new__(cls)
        self.default = default
        self.description = description
        self.emoji = emoji
        self.label = label
        self.value = value
        return self


    @classmethod
    def from_data(cls, data):
        """
        Creates a new string select option from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            String select option data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        
        self.default = parse_default(data)
        self.description = parse_description(data)
        self.emoji = parse_emoji(data)
        self.label = parse_label(data)
        self.value = parse_value(data)
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the string select option to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        put_default(self.default, data, defaults)
        put_description(self.description, data, defaults)
        put_emoji_into(self.emoji, data, defaults)
        put_label(self.label, data, defaults)
        put_value(self.value, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the string select option's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # System fields : value
        
        # value
        repr_parts.append(', value = ')
        repr_parts.append(reprlib.repr(self.value))
        
        # Text fields : emoji & label
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(', emoji = ')
            repr_parts.append(repr(emoji))
        
        # label
        label = self.label
        if (label is not None):
            repr_parts.append(', label = ')
            repr_parts.append(reprlib.repr(label))
        
        # Optional descriptive fields: description & default
        
        # description
        description = self.description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(reprlib.repr(description))
        
        # default
        if self.default:
            repr_parts.append(', default = True')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two string select options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # default
        if self.default != other.default:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # label
        if self.label != other.label:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the string select option's hash value."""
        hash_value = 0
        
        # default
        if self.default:
            hash_value ^= 1 << 8
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= emoji.id
        
        # label
        label = self.label
        if (label is not None):
            hash_value ^= hash(label)
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the string select option.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.default = self.default
        new.description = self.description
        new.emoji = self.emoji
        new.label = self.label
        new.value = self.value
        return new
    
    
    def copy_with(
        self,
        default = ...,
        description = ...,
        emoji = ...,
        label = ...,
        value = ...,
    ):
        """
        Copes the string select with modifying it's defined attributes.
        
        Parameters
        ----------
        default : `bool`, Optional (Keyword only)
            Whether this the the default option. Defaults to `False`.
        
        description : `None`, `str`, Optional (Keyword only)
            Description of the component option.
        
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            Emoji of the option if applicable.
        
        label : `str`, Optional (Keyword only)
            Label of the component option.
        
        value : `str`, Optional (Keyword only)
            The option's value.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # default
        if default is ...:
            default = self.default
        else:
            default = validate_default(default)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
           emoji = validate_emoji(emoji)
        
        
        # label
        if label is ...:
            label = self.label
        else:
            label = validate_label(label)
        
        # value
        if value is ...:
            value = self.value
        else:
            value = validate_value(value)
        
        # Post validation
        if label is None:
            label = value
        
        # Construct
        new = object.__new__(type(self))
        new.default = default
        new.description = description
        new.emoji = emoji
        new.label = label
        new.value = value
        return new



from .plugin_helpers import *
from .plugin_tree import *
from .plugin_tree_helpers import *
from .plugin_tree_iterator import *


__all__ = (
    *plugin_helpers.__all__,
    *plugin_tree.__all__,
    *plugin_tree_helpers.__all__,
    *plugin_tree_iterator.__all__,
)



__all__ = ('EmbeddedActivity',)

from warnings import warn

from ...activity import Activity
from ...bases import DiscordEntity
from ...channel import ChannelType, create_partial_channel_from_id
from ...core import EMBEDDED_ACTIVITIES, GUILDS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_application_id, parse_guild_id, parse_id, parse_launch_id, parse_location, parse_user_states,
    put_application_id, put_guild_id, put_id, put_launch_id, put_location,
    put_user_states, validate_application_id, validate_guild_id, validate_id, validate_launch_id,
    validate_location, validate_user_states
)
from .helpers import _add_embedded_activity_to_guild_cache, _remove_embedded_activity_from_guild_cache


PRECREATE_FIELDS = {
    'application_id': ('application_id', validate_application_id),
    'guild_id': ('guild_id', validate_guild_id),
    'launch_id': ('launch_id', validate_launch_id),
    'location': ('location', validate_location),
    'user_states': ('user_states', validate_user_states),
}


class EmbeddedActivity(DiscordEntity):
    """
    Represents an embedded activity's state.
    
    Attributes
    ----------
    application_id : `int`
        The owner application's identifier.
    
    guild_id : `int`
        The respective guild's identifier.
    
    id : `int`
        The embedded activity's identifier.
    
    launch_id : `int`
        Unique identifier for the activity launch.
    
    location : `None | EmbeddedActivityLocation`
        Where the activity is located.
    
    user_states : `dict<int, EmbeddedActivityUserState>`
        The users state in the activity.
    
    Notes
    -----
    Embedded activities are weakreferable.
    """
    __slots__ = ('__weakref__', 'application_id', 'guild_id', 'launch_id', 'location', 'user_states')
    
    def __new__(
        cls,
        *,
        application_id = ...,
        location = ...,
    ):
        """
        Creates a new partial embedded activities with the given fields.
        
        Parameters
        ----------
        application_id : `int | Application`, Optional (Keyword only)
            The owner application's identifier.
        
        location : `None | EmbeddedActivityLocation`, Optional (Keyword only)
            Where the activity is located.
        
        Raises
        ------
        TypeError
            - If a parameter's value is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # channel_id
        if application_id is ...:
            application_id = 0
        else:
            application_id = validate_application_id(application_id)
        
        # location
        if location is ...:
            location = None
        else:
            location = validate_location(location)
        
        # Construct
        self = object.__new__(cls)
        self.application_id = application_id
        self.guild_id = 0
        self.id = 0
        self.launch_id = 0
        self.location = location
        self.user_states = {}
        return self
    
    
    def __repr__(self):
        """Returns the embedded activity's representation."""
        repr_parts = ['<', type(self).__name__]
        
        embedded_activity_id = self.id
        if embedded_activity_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(embedded_activity_id))
            field_added = True
        
        else:
            repr_parts.append(' (partial)')
            field_added = False
        
        guild_id = self.guild_id
        if guild_id:
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' guild_id = ')
            repr_parts.append(repr(guild_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embedded activities are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id == other_id
        
        return self._is_equal_partial(other)
    
    
    def __ne__(self, other):
        """Returns whether the two embedded activities are not the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id != other_id
        
        return not self._is_equal_partial(other)
    
    
    def _is_equal_partial(self, other):
        """
        Returns whether the embedded activity is equal to the given one.
        This function is called when one or both the embedded activities are partial.
        
        Parameters
        ----------
        other : ``instance<type<self>>``
            The other embedded_activity to compare self to.
        
        Returns
        -------
        is_equal : `bool`
        """
        # application_id
        if self.application_id != other.application_id:
            return False
        
        # guild_id
        # Skip it, non partial field
        
        # id
        # Skip it, non partial field
        
        # launch_id
        # Skip it, non partial field
        
        # location
        if self.location != other.location:
            return False
        
        # user_states
        # Skip it, non partial field
        
        return True
    
    
    def __hash__(self):
        """Returns the embedded activity's hash value."""
        embedded_activity_id = self.id
        if embedded_activity_id:
            return embedded_activity_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns the embedded activity's hash value.
        This function is called when the embedded activity is partial.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # application_id
        hash_value ^= self.application_id
        
        # guild_id
        # Skip it, non partial field
        
        # id
        # Skip it, non partial field
        
        # launch_id
        # Skip it, non partial field
        
        # location
        location = self.location
        if (location is not None):
            hash_value ^= hash(location)
        
        # user_states
        # Skip it, non partial field.
        
        return hash_value
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0):
        """
        Creates a new embedded activities from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Embedded activity data.
        
        guild_id : `int` = `0`, Optional
            The respective guild's identifier.
            
            If given the instance will not be strong cached.
        
        Returns
        -------
        self : `instance<cls>`
        """
        embedded_activity_id = parse_id(data)
        
        try:
            self = EMBEDDED_ACTIVITIES[embedded_activity_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = embedded_activity_id
            self._set_attributes(data, True, guild_id)
            EMBEDDED_ACTIVITIES[embedded_activity_id] = self
        
        else:
            if self.partial:
                self._set_attributes(data, False, guild_id)
        
        return self
    
    
    @classmethod
    def from_data_is_created(cls, data):
        """
        Creates a new embedded activities from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Embedded activity data.
        
        Returns
        -------
        self : `instance<cls>`
        created : `bool`
        """
        embedded_activity_id = parse_id(data)
        
        try:
            self = EMBEDDED_ACTIVITIES[embedded_activity_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = embedded_activity_id
            self._set_attributes(data, True, 0)
            EMBEDDED_ACTIVITIES[embedded_activity_id] = self
            created = True
        
        else:
            if self.partial:
                self._set_attributes(data, False, 0)
                created = True
            
            else:
                created = False
        
        return self, created
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the embedded activities into it's json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_application_id(self.application_id, data, defaults)
        put_location(self.location, data, defaults)
        
        if include_internals:
            put_guild_id(self.guild_id, data, defaults)
            put_id(self.id, data, defaults)
            put_launch_id(self.launch_id, data, defaults)
            put_user_states(self.user_states, data, defaults)
        
        return data
    
    
    def _set_attributes(self, data, creation, guild_id):
        """
        Sets the attributes of the embedded activity.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Embedded activity data.
        
        guild_id : `int` = `0`, Optional
            The respective guild's identifier.
        """
        if guild_id:
            strong_cache = False
        else:
            strong_cache = True
            guild_id = parse_guild_id(data)
        
        self.application_id = parse_application_id(data)
        self.guild_id = guild_id
        self.launch_id = parse_launch_id(data)
        self.location = parse_location(data)
        self.user_states = user_states = parse_user_states(data, ({} if creation else self.user_states), guild_id)
        
        if strong_cache and user_states:
            _add_embedded_activity_to_guild_cache(self)
    
    
    def _update_user_states(self, data):
        """
        Updates the user id-s of the embedded activities.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Embedded activity data.
        """
        user_states = parse_user_states(data, self.user_states, self.guild_id)
        if not user_states:
            _remove_embedded_activity_from_guild_cache(self)
    
    
    def _difference_update_user_states(self, data):
        """
        Difference updates the user id-s of the embedded activities.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Embedded activity data.
        
        Returns
        -------
        joined_users : `set<ClientUserBase>`
            The joined users.
        left_users : `set<ClientUserBase>`
            The left users.
        """
        user_states = self.user_states
        old_users = {user_state.user for user_state in self.user_states.values()}
        parse_user_states(data, self.user_states, self.guild_id)
        new_users = {user_state.user for user_state in self.user_states.values()}
        
        if not user_states:
            _remove_embedded_activity_from_guild_cache(self)
        
        return new_users - old_users, old_users - new_users
    
    
    @classmethod
    def precreate(cls, embedded_activity_id, **keyword_parameters):
        """
        Precreates an embedded activity by creating a partial one with the given parameters.
        
        Parameters
        ----------
        embedded_activity_id : `int | str`
            The embedded activity's identifier.
        
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the role.
        
        Other Parameters
        ----------------
        application_id : `int | str | Application`, Optional (Keyword only)
            The owner application's identifier.
        
        guild_id : `int | str | Guild`, Optional (Keyword only)
            The respective guild's identifier.
        
        launch_id : `int | str`, Optional (Keyword only)
            Unique identifier for the activity launch.
        
        location : `None | EmbeddedActivityLocation`, Optional (Keyword only)
            Where the activity is located.
        
        user_states : `None | iterable<EmbeddedActivityUserState> | dict<int, EmbeddedActivityUserState>` \
                Optional (Keyword only)
            The users state in the activity.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            - If a parameter's value is incorrect
        """
        embedded_activity_id = validate_id(embedded_activity_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = EMBEDDED_ACTIVITIES[embedded_activity_id]
        except KeyError:
            self = cls._create_empty(embedded_activity_id)
            EMBEDDED_ACTIVITIES[embedded_activity_id] = self
        else:
            if (not self.partial):
                return self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, embedded_activity_id):
        """
        Creates an empty embedded activity with the given identifier.
        
        Parameters
        ----------
        embedded_activity_id : `int | str`
            The embedded activity's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.application_id = 0
        self.guild_id = 0
        self.id = embedded_activity_id
        self.launch_id = 0
        self.location = None
        self.user_states = {}
        return self
    
    
    def copy(self):
        """
        Copies the embedded activities.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.application_id = self.application_id
        new.guild_id = 0
        new.id = 0
        new.launch_id = 0
        location = self.location
        if (location is not None):
            location = location.copy()
        new.location = location
        new.user_states = {}
        return new
    
    
    def copy_with(
        self,
        *, 
        application_id = ...,
        location = ...,
    ):
        """
        Copies the embedded activities with the given fields.
        
        
        Parameters
        ----------
        application_id : `int | Application`, Optional (Keyword only)
            The owner application's identifier.
        
        location : `None | EmbeddedActivityLocation`, Optional (Keyword only)
            Where the activity is located.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's value is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # channel_id
        if application_id is ...:
            application_id = self.application_id
        else:
            application_id = validate_application_id(application_id)
        
        # location
        if location is ...:
            location = self.location
            if (location is not None):
                location = location.copy()
        else:
            location = validate_location(location)
        
        # Construct
        new = object.__new__(type(self))
        new.application_id = application_id
        new.guild_id = 0
        new.id = 0
        new.launch_id = 0
        new.location = location
        new.user_states = {}
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the embedded activity state is partial.
        
        Returns
        -------
        partial : `bool`
        """
        guild_id = self.guild_id
        if not guild_id:
            return True
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return True
        
        embedded_activities = guild.embedded_activities
        if embedded_activities is None:
            return True
        
        if self not in embedded_activities:
            return True
        
        return guild.partial
    
    
    @property
    def guild(self):
        """
        Returns the guild where the embedded activity is. The guild must be cached.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        return GUILDS.get(self.guild_id, None)
    
    
    @property
    def channel_id(self):
        """
        Returns the channel's identifier where the embedded activity is.
        
        Returns
        -------
        channel_id : `int`
        """
        location = self.location
        if location is None:
            return 0
        
        return location.channel_id
    
    
    @property
    def channel(self):
        """
        Returns the channel where the embedded activity is.
        
        Returns
        -------
        channel : `None | Channel`
        """
        location = self.location
        if location is None:
            return None
        
        channel_id = location.channel_id
        if not channel_id:
            return None
        
        return create_partial_channel_from_id(channel_id, ChannelType.unknown, self.guild_id)
    
    
    def iter_user_states(self):
        """
        Iterates over the user states of the embedded activity.
        
        This function is an iterable generator.
        
        Yields
        ------
        user_state : ``EmbeddedActivityUserState``
        """
        yield from self.user_states.values()
    
    
    def iter_users(self):
        """
        Iterates over the users in the embedded activity.
        
        This function is an iterable generator.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        for user_state in self.user_states.values():
            yield user_state.user
    
    
    # Deprecated
    
    @property
    def activity(self):
        """
        Deprecated and will be removed in 2025 February.
        """
        warn(
            (
                f'`{type(self).__name__}.activity` is deprecated and will be removed in 2025 February.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return Activity()
    
    
    @property
    def user_ids(self):
        """
        Deprecated and will be removed in 2025 February.
        """
        warn(
            (
                f'`{type(self).__name__}.user_ids` is deprecated and will be removed in 2025 February.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return {user_state.user_id for user_state in self.iter_user_states()}
    

    @property
    def users(self):
        """
        Deprecated and will be removed in 2025 February.
        """
        warn(
            (
                f'`{type(self).__name__}.users` is deprecated and will be removed in 2025 February.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return [user_state.user for user_state in self.iter_user_states()]



__all__ = (
    'ApplicationDiscoverabilityState', 'ApplicationEventWebhookEventType', 'ApplicationEventWebhookState',
    'ApplicationExplicitContentFilterLevel', 'ApplicationIntegrationType', 'ApplicationInteractionEventType',
    'ApplicationInteractionVersion', 'ApplicationInternalGuildRestriction', 'ApplicationMonetizationState',
    'ApplicationRPCState', 'ApplicationStoreState', 'ApplicationType', 'ApplicationVerificationState'
)

from ...bases import Preinstance as P, PreinstancedBase


class ApplicationDiscoverabilityState(PreinstancedBase, value_type = int):
    """
    Represents an application's discoverability state.
    
    Attributes
    ----------
    name : `str`
        The name of the application discoverability state.
    
    value : `int`
        The Discord side identifier value of the application discoverability state.
        
    Type Attributes
    ---------------
    Every predefined application discoverability state can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | ineligible                | ineligible                | 1     |
    +---------------------------+---------------------------+-------+
    | not_discoverable          | not discoverable          | 2     |
    +---------------------------+---------------------------+-------+
    | discoverable              | discoverable              | 3     |
    +---------------------------+---------------------------+-------+
    | featurable                | featurable                | 4     |
    +---------------------------+---------------------------+-------+
    | blocked                   | blocked                   | 5     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    ineligible = P(1, 'ineligible')
    not_discoverable = P(2, 'not discoverable')
    discoverable = P(3, 'discoverable')
    featurable = P(4, 'featurable')
    blocked = P(5, 'blocked')


class ApplicationEventWebhookEventType(PreinstancedBase, value_type = str):
    """
    Represents an application's event webhook event type.
    
    Attributes
    ----------
    name : `str`
        The name of the application event webhook event type.
    
    value : `str`
        The Discord side identifier value of the application event webhook event type.
        
    Type Attributes
    ---------------
    Every predefined application event webhook event type can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+---------------------------+
    | Type attribute name       | name                      | value                     |
    +===========================+===========================+===========================+
    | application_authorization | application authorization | 'APPLICATION_AUTHORIZED'  |
    +---------------------------+---------------------------+---------------------------+
    | entitlement_create        | entitlement create        | 'ENTITLEMENT_CREATE'      |
    +---------------------------+---------------------------+---------------------------+
    | quest_enrollment          | quest enrollment          | 'QUEST_USER_ENROLLMENT'   |
    +---------------------------+---------------------------+---------------------------+
    """
    __slots__ = ()
    
    # predefined
    application_authorization = P('APPLICATION_AUTHORIZED', 'application authorization')
    entitlement_create = P('ENTITLEMENT_CREATE', 'entitlement create')
    quest_enrollment = P('QUEST_USER_ENROLLMENT', 'quest enrollment')


class ApplicationEventWebhookState(PreinstancedBase, value_type = int):
    """
    Represents an application's event webhook state.
    
    Attributes
    ----------
    name : `str`
        The name of the application event webhook state.
    
    value : `int`
        The Discord side identifier value of the application event webhook state.
        
    Type Attributes
    ---------------
    Every predefined application event webhook state can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | disabled                  | disabled                  | 1     |
    +---------------------------+---------------------------+-------+
    | enabled                   | enabled                   | 2     |
    +---------------------------+---------------------------+-------+
    | auto_disabled             | auto-disabled             | 3     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    disabled = P(1, 'disabled')
    enabled = P(2, 'enabled')
    auto_disabled = P(3, 'auto-disabled')


class ApplicationExplicitContentFilterLevel(PreinstancedBase, value_type = int):
    """
    Represents an application's explicit content filter level.
    
    Attributes
    ----------
    name : `str`
        The name of the application explicit content filter level.
    
    value : `int`
        The Discord side identifier value of the application explicit content filter level.
        
    Type Attributes
    ---------------
    Every predefined application explicit content filter level can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | filtered                  | filtered                  | 1     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    filtered = P(1, 'filtered')


class ApplicationIntegrationType(PreinstancedBase, value_type = int):
    """
    Represents how an application can be integrated.
    
    Attributes
    ----------
    name : `str`
        The name of the application integration type.
    
    value : `int`
        The Discord side identifier value of the application integration type.
        
    Type Attributes
    ---------------
    Every predefined application integration type can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | guild_install             | guild install             | 0     |
    +---------------------------+---------------------------+-------+
    | user_install              | user install              | 1     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    guild_install = P(0, 'guild install')
    user_install = P(1, 'user install')


class ApplicationInteractionEventType(PreinstancedBase, value_type = str):
    """
    Represents an application's interaction event type.
    
    Attributes
    ----------
    name : `str`
        The name of the application interaction event type.
    
    value : `str`
        The Discord side identifier value of the application interaction event type.
        
    Type Attributes
    ---------------
    Every predefined application interaction event type can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | `''`  |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P('', 'none')


class ApplicationInteractionVersion(PreinstancedBase, value_type = int):
    """
    Represents an application interaction version.
    
    Attributes
    ----------
    name : `str`
        The name of the application interaction version.
    
    value : `int`
        The Discord side identifier value of the application interaction version.
    
    Type Attributes
    ---------------
    Every predefined application interaction version can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | every                     | every                     | 1     |
    +---------------------------+---------------------------+-------+
    | selective                 | selective                 | 2     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    every = P(1, 'every')
    selective = P(2, 'selective')


class ApplicationInternalGuildRestriction(PreinstancedBase, value_type = int):
    """
    Represents an application's internal guild restriction.
    
    > Note that could not find anything about this.
    
    Attributes
    ----------
    name : `str`
        The name of the application internal guild restriction.
    
    value : `int`
        The Discord side identifier value of the application internal guild restriction.
    
    Type Attributes
    ---------------
    Every predefined application internal guild restriction can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | restricted                | restricted                | 1     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    restricted = P(1, 'restricted')


class ApplicationMonetizationState(PreinstancedBase, value_type = int):
    """
    Represents an application's monetization state.
    
    Attributes
    ----------
    enabled : `bool`
        Whether monetization is enabled.
    
    locked : `bool`
        Whether monetization actions are locked.
    
    name : `str`
        The default name of the monetization state.
    
    settable : `bool`
        Whether the monetization state can be set.
    
    value : `int`
        The Discord side identifier value of the monetization state.
    
    Type Attributes
    ---------------
    Every predefined monetization state is also stored as a type attribute:
    
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | Type attribute name   | Name                  | Value     | Enabled   | Locked    | Settable  |
    +=======================+=======================+===========+===========+===========+===========+
    | none                  | none                  | 0         | `False`   | `False`   | `False`   |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | disabled              | disabled              | 1         | `False`   | `False`   | `True`    |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | provisional           | provisional           | 2         | `True`    | `False`   | `True`    |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | rejected              | rejected              | 3         | `False`   | `True`    | `False`   |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | approved              | approved              | 4         | `True`    | `False`   | `False`   |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | blocked               | blocked               | 5         | `False`   | `True`    | `False`   |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    """
    __slots__ = ('enabled', 'locked', 'settable')
    
    def __new__(cls, value, name = None, enabled = False, locked = False, settable = False):
        """
        Creates a monetization state.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the monetization state.
        
        name : `None | str` = `None`, Optional
            The default name of the monetization state.
        
        enabled : `bool` = `False`, Optional
            Whether monetization is enabled.
        
        locked : `bool` = `False`, Optional
            Whether monetization actions are locked.
        
        settable : `bool` = `False`, Optional
            Whether the monetization state can be set.
        """
        self = PreinstancedBase.__new__(cls, value, name)
        self.enabled = enabled
        self.locked = locked
        self.settable = settable
        return self
    
    # predefined
    none = P(0, 'none', False, False, False)
    disabled = P(1, 'disabled', False, False, True)
    provisional = P(2, 'provisional', True, False, True)
    rejected = P(3, 'rejected', False, True, False)
    approved = P(4, 'approved', True, False, False)
    blocked = P(5, 'blocked', False, True, False)


class ApplicationRPCState(PreinstancedBase, value_type = int):
    """
    Represents an application rpc state.
    
    Attributes
    ----------
    name : `str`
        The name of the application rpc state.
    
    value : `int`
        The Discord side identifier value of the application rpc state.
    
    Type Attributes
    ---------------
    Every predefined application rpc state can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | not_submitted             | not submitted             | 1     |
    +---------------------------+---------------------------+-------+
    | submitted                 | submitted                 | 2     |
    +---------------------------+---------------------------+-------+
    | approved                  | approved                  | 3     |
    +---------------------------+---------------------------+-------+
    | rejected                  | rejected                  | 4     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    not_submitted = P(1, 'not submitted')
    submitted = P(2, 'submitted')
    approved = P(3, 'approved')
    rejected = P(4, 'rejected')


class ApplicationStoreState(PreinstancedBase, value_type = int):
    """
    Represents an application store state.
    
    Attributes
    ----------
    name : `str`
        The name of the application store state.
    
    value : `int`
        The Discord side identifier value of the application store state.
    
    Type Attributes
    ---------------
    Every predefined application store state can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | no_license                | no license                | 1     |
    +---------------------------+---------------------------+-------+
    | license                   | license                   | 2     |
    +---------------------------+---------------------------+-------+
    | submitted                 | submitted                 | 3     |
    +---------------------------+---------------------------+-------+
    | approved                  | approved                  | 4     |
    +---------------------------+---------------------------+-------+
    | rejected                  | rejected                  | 5     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    no_license = P(1, 'no license')
    license = P(2, 'license')
    submitted = P(3, 'submitted')
    approved = P(4, 'approved')
    rejected = P(5, 'rejected')


class ApplicationType(PreinstancedBase, value_type = int):
    """
    Represents an application type.
    
    Attributes
    ----------
    name : `str`
        The name of the application type.
    
    value : `int`
        The Discord side identifier value of the application type.
        
    Type Attributes
    ---------------
    Every predefined application type can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | game                      | game                      | 1     |
    +---------------------------+---------------------------+-------+
    | music                     | music                     | 2     |
    +---------------------------+---------------------------+-------+
    | ticketed_event            | ticketed event            | 3     |
    +---------------------------+---------------------------+-------+
    | guild_role_subscription   | guild role subscription   | 4     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    game = P(1, 'game')
    music = P(2, 'music')
    ticketed_event = P(3, 'ticketed event')
    guild_role_subscription = P(4, 'guild role subscription')


class ApplicationVerificationState(PreinstancedBase, value_type = int):
    """
    Represents an application's verification state.
    
    Attributes
    ----------
    name : `str`
        The name of the application verification state.
    
    value : `int`
        The Discord side identifier value of the application verification state.
    
    Type Attributes
    ---------------
    Every predefined application verification state can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | ineligible                | ineligible                | 1     |
    +---------------------------+---------------------------+-------+
    | not_submitted             | not submitted             | 2     |
    +---------------------------+---------------------------+-------+
    | submitted                 | submitted                 | 3     |
    +---------------------------+---------------------------+-------+
    | approved                  | approved                  | 4     |
    +---------------------------+---------------------------+-------+
    | blocked                   | blocked                   | 5     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    ineligible = P(1, 'ineligible')
    not_submitted = P(2, 'not submitted')
    submitted = P(3, 'submitted')
    approved = P(4, 'approved')
    blocked = P(5, 'blocked')



__all__ = ('ApplicationCommandOptionMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder

from .constants import MAX_LENGTH_DEFAULT, MIN_LENGTH_DEFAULT


class ApplicationCommandOptionMetadataBase(RichAttributeErrorBaseType):
    """
    Base type for application command metadatas.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new application command option metadata with the given parameters.
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates a new application command option metadata with the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters passed to ``ApplicationCommandOption.__new__``.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        return cls()
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new application command option metadata from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application command option data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the application command option metadata to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        return {}
    
    
    def __repr__(self):
        """Returns the application command option metadata's representation."""
        repr_parts = ['<', self.__class__.__name__]
        self._add_type_specific_repr_fields(repr_parts)
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _add_type_specific_repr_fields(self, repr_parts):
        """
        Adds the type specific fields into the given into the representation parts.
        
        Parameters
        ----------
        repr_parts : `list` of `str`
        """
    
    def __eq__(self, other):
        """Returns whether the two application command options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two application command options are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be the same type as `self`.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the application command option metadata."""
        return 0
    
    
    def copy(self):
        """
        Copies the application command option metadata.
        
        Returns
        -------
        new : `instance<type<cls>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the application command option metadata with the given fields.
        
        Returns
        -------
        new : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        return self.copy()
    
    
    def copy_with_keyword_parameters(self, keyword_parameters):
        """
        Copies the application command option metadata with the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters passed to ``ApplicationCommandOption.copy_with``.
        
        Returns
        -------
        new : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        return self.copy_with()
    
    
    autocomplete = PlaceHolder(
        False,
        """
        Whether the option supports auto completion.
        
        Applicable for string, integer and float options.
        
        Returns
        -------
        autocomplete : `bool`
        """
    )
    
    
    channel_types = PlaceHolder(
        None,
        """
        The accepted channel types by the option.
        
        Applicable for channel options.
        
        Returns
        -------
        channel_types : `None`, `tuple` of ``ChannelType``
        """
    )
    
    
    choices = PlaceHolder(
        None,
        """
        Choices for `str` and `int` types for the user to pick from.
        
        Applicable for string, integer and float options.
        
        Returns
        -------
        choices : `None`, `tuple` of ``ApplicationCommandOptionChoice``
        """
    )
    
    
    default = PlaceHolder(
        False,
        """
        Whether the option is the default one.
        
        Applicable for sub-command options.
        
        Returns
        -------
        default : `bool`
        """
    )
    
    
    max_length = PlaceHolder(
        MAX_LENGTH_DEFAULT,
        """
        The maximum input length allowed for this option.
        
        Applicable for string options.
        
        Returns
        -------
        max_length : `int`
        """
    )
    
    
    max_value = PlaceHolder(
        None,
        """
        The maximal value permitted for this option.
        
        Applicable for integer and float options.
        
        Returns
        -------
        max_value : `None`, `int`, `float`
        """
    )
    
    
    min_length = PlaceHolder(
        MIN_LENGTH_DEFAULT,
        """
        The minimal input length allowed for this option.
        
        Applicable for string options.
        
        Returns
        -------
        min_length : `int`
        """
    )
    
    
    min_value = PlaceHolder(
        None,
        """
        The minimal value permitted for this option.
        
        Applicable for integer and float options.
        
        Returns
        -------
        min_value : `None`, `int`, `float`
        """
    )
    
    
    options = PlaceHolder(
        None,
        """
        Contains the option's parameter or its sub-commands (or groups).
        
        Applicable for sub-command and sub-command group options.
        
        Returns
        -------
        options : `None`, `tuple` of ``ApplicationCommandOption``
        """
    )
    
    
    required = PlaceHolder(
        False,
        """
        Whether the parameter is required.
        
        Applicable for all parameter option.
        
        Returns
        -------
        required : `bool`
        """
    )



__all__ = ('create_partial_invite_data', 'create_partial_invite_from_data')

from ...core import INVITES

from .fields import parse_code, put_channel_id, put_code, put_guild_id,\
    validate_max_age, put_max_age, validate_max_uses, put_temporary, \
    put_max_uses, validate_temporary, validate_unique, put_unique, \
    validate_target_type, validate_target_application_id, put_target_application_id, \
    put_target_user_id, put_target_type, validate_target_user_id

from .invite import Invite


INVITE_GUILD_FIELD_CONVERTERS = {
    'max_age': (validate_max_age, put_max_age),
    'max_uses': (validate_max_uses, put_max_uses),
    'target_application': (validate_target_application_id, put_target_application_id),
    'target_application_id': (validate_target_application_id, put_target_application_id),
    'target_type': (validate_target_type, put_target_type),
    'target_user': (validate_target_user_id, put_target_user_id),
    'target_user_id': (validate_target_user_id, put_target_user_id),
    'temporary': (validate_temporary, put_temporary),
    'unique': (validate_unique, put_unique),
}


def create_partial_invite_from_data(data):
    """
    Creates a partial invite from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Partial invite data.
    
    Returns
    -------
    invite : ``Invite``
    """
    invite_code = parse_code(data)
    
    try:
        invite = INVITES[invite_code]
    except KeyError:
        invite = Invite._create_empty(invite_code)
        invite._update_attributes_partial(data)
        INVITES[invite_code] = invite
    else:
        invite._update_attributes_partial(data)
    
    return invite


def create_partial_invite_data(invite):
    """
    Creates partial invite data.
    
    Parameters
    ----------
    invite : ``Invite``
        Invite to create data from.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data = {}
    put_channel_id(invite.channel_id, data, True)
    put_code(invite.code, data, True)
    put_guild_id(invite.guild_id, data, True)
    return data



__all__ = (
    'CONTEXT_TARGET_TYPES', 'INTEGRATION_CONTEXT_TYPES_ALL', 'ApplicationCommandHandlerType',
    'ApplicationCommandIntegrationContextType', 'ApplicationCommandTargetType',
)

from warnings import warn

from scarletio import class_property

from ...bases import Preinstance as P, PreinstancedBase


class ApplicationCommandHandlerType(PreinstancedBase, value_type = int):
    """
    Represents what handles an application command.
    
    Attributes
    ----------
    name : `str`
        The name of the handler.
    
    value : `int`
        The identifier value representing the handler.
    
    Type Attributes
    ---------------
    Every predefined application command handler type can be accessed as type attribute as well:
    
    +---------------------------------------+---------------------------------------+-------+
    | Type attribute name                   | Name                                  | Value |
    +=======================================+=======================================+=======+
    | none                                  | none                                  | 0     |
    +---------------------------------------+---------------------------------------+-------+
    | application                           | application                           | 1     |
    +---------------------------------------+---------------------------------------+-------+
    | discord_embedded_activity_launcher    | discord embedded activity launcher    | 2     |
    +---------------------------------------+---------------------------------------+-------+
    """
    __slots__ = ()
    
    none = P(0, 'none')
    application = P(1, 'application')
    discord_embedded_activity_launcher = P(2, 'discord embedded activity launcher')


class ApplicationCommandIntegrationContextType(PreinstancedBase, value_type = int):
    """
    Represents an application command's integration's context's type or in other words, where it can be used.
    
    Attributes
    ----------
    name : `str`
        The name of the application command integration context.
    
    value : `int`
        The identifier value the application command integration context.
    
    Type Attributes
    ---------------
    Every predefined application command integration context can be accessed as type attribute as well:
    
    +-----------------------+-------------------+-------+
    | Type attribute name   | Name              | Value |
    +=======================+===================+=======+
    | guild                 | guild             | 0     |
    +-----------------------+-------------------+-------+
    | bot_private_channel   | chat              | 1     |
    +-----------------------+-------------------+-------+
    | any_private_channel   | user              | 2     |
    +-----------------------+-------------------+-------+
    """
    __slots__ = ()
    
    guild = P(0, 'guild')
    bot_private_channel = P(1, 'bot private channel')
    any_private_channel = P(2, 'any private channel')


INTEGRATION_CONTEXT_TYPES_ALL = tuple(sorted(
    ApplicationCommandIntegrationContextType.INSTANCES.values()
))


class ApplicationCommandTargetType(PreinstancedBase, value_type = int):
    """
    Represents an application command's target.
    
    Attributes
    ----------
    name : `str`
        The name of the application command target.
    
    value : `int`
        The identifier value the application command target.
    
    Type Attributes
    ---------------
    Every predefined application command target can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | Name                      | Value |
    +============================+==========================+======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | chat                      | chat                      | 1     |
    +---------------------------+---------------------------+-------+
    | user                      | user                      | 2     |
    +---------------------------+---------------------------+-------+
    | message                   | message                   | 3     |
    +---------------------------+---------------------------+-------+
    | embedded_activity_launch  | embedded activity launch  | 4     |
    +---------------------------+---------------------------+-------+
    | channel                   | channel                   | 1004  |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    none = P(0, 'none',)
    chat = P(1, 'chat',)
    user = P(2, 'user',)
    message = P(3, 'message',)
    embedded_activity_launch = P(4, 'embedded activity launch')
    channel = P(1004, 'channel')
    
    @class_property
    def activity_start(cls):
        """
        Deprecated and will be removed in 2025 Jun. Use `.embedded_activity_launch` instead.
        """
        warn(
            (
                f'`{cls.__name__}.activity_start` is deprecated and will be removed in 2025 Jun. '
                f'Use `.embedded_activity_launch` instead.'
            ),
            FutureWarning,
            stacklevel = 3,
        )
        return cls.embedded_activity_launch


CONTEXT_TARGET_TYPES = frozenset((
    ApplicationCommandTargetType.user,
    ApplicationCommandTargetType.message,
    ApplicationCommandTargetType.channel,
))



__all__ = ('ScheduledEventEntityMetadataLocation',)

from scarletio import copy_docs

from .base import ScheduledEventEntityMetadataBase
from .fields import parse_location, put_location, validate_location


class ScheduledEventEntityMetadataLocation(ScheduledEventEntityMetadataBase):
    """
    Location entity metadata of ``ScheduledEvent``-s.
    
    Attributes
    ----------
    location : `None`, `str`
        The place where the event will take place.
    """
    __slots__ = ('location', )
    
    
    def __new__(cls, *, location = ...):
        """
        Creates a new entity metadata instance.
        
        Parameters
        ----------
        location : `None`, `str`, Optional (Keyword only)
            The place where the event will take place.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # location
        if location is ...:
            location = None
        else:
            location = validate_location(location)
        
        self = object.__new__(cls)
        self.location = location
        return self
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            location = keyword_parameters.pop('location', ...),
        )
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.location = parse_location(data)
        return self
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_location(self.location, data, defaults)
        return data
    
    
    @classmethod
    def _create_empty(cls):
        self = object.__new__(cls)
        self.location = None
        return self
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        location = self.location
        if (location is not None):
            hash_value ^= hash(location)
        
        return hash_value
    
    
    @copy_docs(ScheduledEventEntityMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if self.location != other.location:
            return False
        
        return True
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.location = self.location
        return new
    
    
    def copy_with(self, *, location = ...):
        """
        Copies the scheduled event entity metadata with the given fields.
        
        Parameters
        ----------
        location : `None`, `str`, Optional (Keyword only)
            The place where the event will take place.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # location
        if location is ...:
            location = self.location
        else:
            location = validate_location(location)
        
        new = object.__new__(type(self))
        new.location = location
        return new

    
    @copy_docs(ScheduledEventEntityMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            location = keyword_parameters.pop('location', ...),
        )



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .preinstanced import GuildFeature


class GuildPremiumPerks(RichAttributeErrorBaseType):
    """
    Represents the premium perk that a user gets for each boost tier.
    
    Attributes
    ----------
    bitrate_limit : `int`
        The maximal bit rate usable in voice channel.
    concurrent_activities : `int`
        The maximal amount of concurrent embedded activities. Set as `0` if unlimited.
    emoji_limit : `int`
        The maximal amount of emojis.
    features : `None`, `tuple` of ``GuildFeature``
        The guild features acquired with the premium tier.
    screen_share_frame_rate : `int`
        The maximal frame rate usable when screen sharing.
    screen_share_resolution : `str`
        The maximal screen resolution's name when screen sharing.
    sound_limit : `int`
        The maximal amount of sound board sounds.
    sticker_limit : `int`
        The maximal amount of stickers the guild can have.
    tier : `int`
        The premium tier's level.
    upload_limit : `int`
        The maximal upload limit in bytes.
    """
    __slots__ = (
        'bitrate_limit', 'concurrent_activities', 'emoji_limit', 'features', 'screen_share_frame_rate',
        'screen_share_resolution', 'sound_limit', 'sticker_limit', 'tier', 'upload_limit', 
    )
    
    def __new__(
        cls,
        bitrate_limit,
        concurrent_activities,
        emoji_limit,
        features,
        screen_share_frame_rate,
        screen_share_resolution,
        sound_limit,
        sticker_limit,
        tier,
        upload_limit,
    ):
        """
        Creates a new guild premium tier perk.
        
        Parameters
        ----------
        bitrate_limit : `int`
            The maximal bit rate usable in voice channel.
        concurrent_activities : `int`
            The maximal amount of concurrent embedded activities. Set as `0` if unlimited.
        emoji_limit : `int`
            The maximal amount of emojis.
        features : `None`, `tuple` of ``GuildFeature``
            The guild features acquired with the premium tier.
        screen_share_frame_rate : `int`
            The maximal frame rate usable when screen sharing.
        screen_share_resolution : `str`
            The maximal screen resolution's name when screen sharing.
        sound_limit : `int`
            The maximal amount of sound board sounds.
        sticker_limit : `int`
            The maximal amount of stickers the guild can have.
        tier : `int`
            The premium tier's level.
        upload_limit : `int`
            The maximal upload limit in bytes.
        """
        self = object.__new__(cls)
        self.bitrate_limit = bitrate_limit
        self.concurrent_activities = concurrent_activities
        self.emoji_limit = emoji_limit
        self.features = features
        self.screen_share_frame_rate = screen_share_frame_rate
        self.screen_share_resolution = screen_share_resolution
        self.sound_limit = sound_limit
        self.sticker_limit = sticker_limit
        self.tier = tier
        self.upload_limit = upload_limit
        return self
    
    
    def __repr__(self):
        """Returns the guild premium perk's representation."""
        return f'{self.__class__.__name__} tier = {self.tier!r}>'
    
    
    def iter_features(self):
        """
        Iterates over the features of the guild premium perk.
        
        This method is an iterable generator.
        
        Yields
        ------
        feature : ``GuildFeature``
        """
        features = self.features
        if (features is not None):
            yield from features


TIER_0 = GuildPremiumPerks(
    96000,
    2,
    50,
    None,
    30,
    '720p',
    8,
    5,
    0,
    8388608,
)


TIER_1 = GuildPremiumPerks(
    128000,
    3,
    100,
    (
        GuildFeature.animated_icon,
        GuildFeature.invite_splash,
    ),
    60,
    '720p',
    24,
    15,
    1,
    26214400,
)


TIER_2 = GuildPremiumPerks(
    256000,
    5,
    150,
    (
        GuildFeature.animated_icon,
        GuildFeature.banner,
        GuildFeature.invite_splash,
        GuildFeature.role_icons,
    ),
    60,
    '1080p',
    36,
    30,
    2,
    52428800,
)


TIER_3 = GuildPremiumPerks(
    384000,
    0,
    250,
    (
        GuildFeature.animated_banner,
        GuildFeature.animated_icon,
        GuildFeature.banner,
        GuildFeature.invite_splash,
        GuildFeature.role_icons,
        GuildFeature.vanity_invite,
    ),
    60,
    '1080p',
    48,
    60,
    3,
    104857600,
)


TIERS = {
    0: TIER_0,
    1: TIER_1,
    2: TIER_2,
    3: TIER_3,
}

TIER_MAX = TIER_3



__all__ = ('ChannelTypeFlag', )

from scarletio import class_property

from ...bases import FlagBase, FlagDescriptor as F


CHANNEL_TYPE_SHIFT_CONNECTABLE = 0
CHANNEL_TYPE_SHIFT_GUILD = 1
CHANNEL_TYPE_SHIFT_GUILD_SORTABLE = 2
CHANNEL_TYPE_SHIFT_GUILD_SYSTEM = 3
CHANNEL_TYPE_SHIFT_INVITABLE = 4
CHANNEL_TYPE_SHIFT_PRIVATE = 5
CHANNEL_TYPE_SHIFT_READABLE = 6
CHANNEL_TYPE_SHIFT_TEXTUAL = 7
CHANNEL_TYPE_SHIFT_THREAD = 8
CHANNEL_TYPE_SHIFT_THREADABLE = 9
CHANNEL_TYPE_SHIFT_FORUM = 10

CHANNEL_TYPE_MASK_CONNECTABLE = 1 << CHANNEL_TYPE_SHIFT_CONNECTABLE
CHANNEL_TYPE_MASK_GUILD = 1 << CHANNEL_TYPE_SHIFT_GUILD
CHANNEL_TYPE_MASK_GUILD_SORTABLE = 1 << CHANNEL_TYPE_SHIFT_GUILD_SORTABLE
CHANNEL_TYPE_MASK_GUILD_SYSTEM = 1 << CHANNEL_TYPE_SHIFT_GUILD_SYSTEM
CHANNEL_TYPE_MASK_INVITABLE = 1 << CHANNEL_TYPE_SHIFT_INVITABLE
CHANNEL_TYPE_MASK_PRIVATE = 1 << CHANNEL_TYPE_SHIFT_PRIVATE
CHANNEL_TYPE_MASK_READABLE = 1 << CHANNEL_TYPE_SHIFT_READABLE
CHANNEL_TYPE_MASK_TEXTUAL = 1 << CHANNEL_TYPE_SHIFT_TEXTUAL
CHANNEL_TYPE_MASK_THREAD = 1 << CHANNEL_TYPE_SHIFT_THREAD
CHANNEL_TYPE_MASK_THREADABLE = 1 << CHANNEL_TYPE_SHIFT_THREADABLE
CHANNEL_TYPE_MASK_FORUM = 1 << CHANNEL_TYPE_SHIFT_FORUM


class ChannelTypeFlag(FlagBase):
    """
    Represents a channel type's flags.
    
    Channel type flags are used to define for which role a channel type can be used.
    
    The implemented channel flags are the following:
    
    +-------------------------------+-------------------+
    | Respective name               | Bitwise position  |
    +===============================+===================+
    | connectable                   | 0                 |
    +-------------------------------+-------------------+
    | guild                         | 1                 |
    +-------------------------------+-------------------+
    | guild_sortable                | 2                 |
    +-------------------------------+-------------------+
    | guild_system                  | 3                 |
    +-------------------------------+-------------------+
    | invitable                     | 4                 |
    +-------------------------------+-------------------+
    | private                       | 5                 |
    +-------------------------------+-------------------+
    | readable                      | 6                 |
    +-------------------------------+-------------------+
    | textual                       | 7                 |
    +-------------------------------+-------------------+
    | thread                        | 8                 |
    +-------------------------------+-------------------+
    | threadable                    | 9                 |
    +-------------------------------+-------------------+
    | forum                         | 10                |
    +-------------------------------+-------------------+
    """
    connectable = F(CHANNEL_TYPE_SHIFT_CONNECTABLE)
    guild = F(CHANNEL_TYPE_SHIFT_GUILD)
    guild_sortable = F(CHANNEL_TYPE_SHIFT_GUILD_SORTABLE)
    guild_system = F(CHANNEL_TYPE_SHIFT_GUILD_SYSTEM)
    invitable = F(CHANNEL_TYPE_SHIFT_INVITABLE)
    private = F(CHANNEL_TYPE_SHIFT_PRIVATE)
    readable = F(CHANNEL_TYPE_SHIFT_READABLE)
    textual = F(CHANNEL_TYPE_SHIFT_TEXTUAL)
    thread = F(CHANNEL_TYPE_SHIFT_THREAD)
    threadable = F(CHANNEL_TYPE_SHIFT_THREADABLE)
    forum = F(CHANNEL_TYPE_SHIFT_FORUM)
    
    
    @class_property
    def all(cls):
        """
        Returns a value which has all flags.
        
        Returns
        -------
        value : ``ChannelTypeFlag``
        """
        value = 0
        
        for shift in cls.__shifts__.values():
            value |= 1 << shift
        
        return cls(value)



__all__ = ()

from scarletio import include

from ...bases import maybe_snowflake
from ...field_parsers import entity_id_parser_factory, flag_parser_factory
from ...field_putters import entity_id_putter_factory, preinstanced_putter_factory, string_flag_putter_factory
from ...field_validators import entity_id_validator_factory, flag_validator_factory, preinstanced_validator_factory
from ...permission import Permission
from ...permission.constants import PERMISSION_ALLOW_KEY, PERMISSION_DENY_KEY

from .helpers import get_permission_overwrite_key_value
from .preinstanced import PermissionOverwriteTargetType


Role = include('Role')
ClientUserBase = include('ClientUserBase')

# allow

parse_allow = flag_parser_factory(PERMISSION_ALLOW_KEY, Permission)
put_allow = string_flag_putter_factory(PERMISSION_ALLOW_KEY)
validate_allow = flag_validator_factory('allow', Permission)

# deny

parse_deny = flag_parser_factory(PERMISSION_DENY_KEY, Permission)
put_deny = string_flag_putter_factory(PERMISSION_DENY_KEY)
validate_deny = flag_validator_factory('deny', Permission)

# target

def put_target(target, data, defaults, *, include_internals = False):
    """
    Puts the `target` field of ``PermissionOverwrite`` to the given data.
    
    The `target` value must come from ``validate_target``.
    
    Parameters
    ----------
    target : `tuple` (`int`, ``PermissionOverwriteTargetType``)
        Permission overwrite `target-id`, `target_type` pair.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    include_internals : `bool` = `False`, Optional (Keyword only)
        Whether internal fields should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    target_id, target_type = target
    
    if include_internals:
        data['id'] = str(target_id) if target_id else None
    
    if (target_type is not PermissionOverwriteTargetType.unknown):
        data['type'] = target_type.value
    
    return data


def validate_target(target):
    """
    Validates the `target` field of ``PermissionOverwrite``. target is a mixed representation of `target_id` and
    `target_type`.
    
    Parameters
    ----------
    target : ``Role``, ``ClientUserBase``, `int`
        The permission overwrite's target or it's identifier.
    
    Returns
    -------
    target_id : `int`
        The permission overwrite target's identifier.
    target_type : ``PermissionOverwriteTargetType``
        The permission overwrite's type.
    
    Raises
    ------
    TypeError
        - If `target`'s type is unexpected.
    """
    # target_id
    if isinstance(target, Role):
        target_id = target.id
        target_type = PermissionOverwriteTargetType.role
    
    elif isinstance(target, ClientUserBase):
        target_id = target.id
        target_type = PermissionOverwriteTargetType.user
    
    else:
        target_id = maybe_snowflake(target)
        if (target_id is None):
            raise TypeError(
                f'`target` can be `int`, `{Role.__name__}`, `{ClientUserBase.__name__}`, got '
                f'{target.__class__.__name__}; {target!r}.'
            )
        
        target_type = PermissionOverwriteTargetType.unknown
    
    return target_id, target_type

# target_id

parse_target_id = entity_id_parser_factory('id')
put_target_id = entity_id_putter_factory('id')
validate_target_id = entity_id_validator_factory('target_id')

# target_type

def parse_target_type(data):
    """
    Parses out ``PermissionOverwrite.target_type`` from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Permission overwrite data.
    
    Returns
    -------
    target_type : ``PermissionOverwriteTargetType``
    """
    return PermissionOverwriteTargetType(get_permission_overwrite_key_value(data))


put_target_type = preinstanced_putter_factory('type')
validate_target_type = preinstanced_validator_factory('target_type', PermissionOverwriteTargetType)



__all__ = ()

from ...integration import IntegrationType
from ...integration.integration.fields import validate_type as validate_integration_type
from ...user.voice_state.fields import validate_channel_id

from ..audit_log_entry_detail_conversion import AuditLogEntryDetailConversion, AuditLogEntryDetailConversionGroup
from ..conversion_helpers.converters import value_deserializer_id, value_serializer_id


# ---- channel_id ----

CHANNEL_ID_CONVERSION = AuditLogEntryDetailConversion(
    'channel_id',
    'channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_channel_id,
)


# ---- count ----

COUNT_CONVERSION = AuditLogEntryDetailConversion(
    'count',
    'count',
)

@COUNT_CONVERSION.set_value_deserializer
def count_value_deserializer(value):
    if value is None:
        value = 0
    return value


@COUNT_CONVERSION.set_value_validator
def validate_count(value):
    if not isinstance(value, int):
        raise TypeError(f'`count` can be `int`, got {value.__class__.__name__}; {value!r}.')
    
    if value < 0:
        raise ValueError(f'`count` cannot be negative, got {value!r}.')
    
    return value



# ---- delete_message_duration ----

DELETE_MESSAGE_DURATION_CONVERSION = AuditLogEntryDetailConversion(
    'delete_message_seconds',
    'delete_message_duration',
)


@DELETE_MESSAGE_DURATION_CONVERSION.set_value_deserializer
def delete_message_duration_value_deserializer(value):
    if value is None:
        value = 0
    return value


@DELETE_MESSAGE_DURATION_CONVERSION.set_value_validator
def validate_delete_message_duration(value):
    if not isinstance(value, int):
        raise TypeError(f'`delete_message_duration` can be `int`, got {value.__class__.__name__}; {value!r}.')
    
    if value < 0:
        raise ValueError(f'`delete_message_duration` cannot be negative, got {value!r}.')
    
    return value


# ---- delete_message_duration | deprecated version ----

DELETE_MESSAGE_DURATION_DEPRECATED_DAYS_CONVERSION = AuditLogEntryDetailConversion(
    'delete_message_days',
    'delete_message_duration',
)


@DELETE_MESSAGE_DURATION_DEPRECATED_DAYS_CONVERSION.set_value_deserializer
def delete_message_days_value_deserializer(value):
    return value * 24 * 60 * 60


# ---- integration_type ----

INTEGRATION_TYPE_CONVERSION = AuditLogEntryDetailConversion(
    'integration_type', 'integration_type', value_validator = validate_integration_type
)


@INTEGRATION_TYPE_CONVERSION.set_value_deserializer
def integration_type_value_deserializer(value):
    return IntegrationType(value)


@INTEGRATION_TYPE_CONVERSION.set_value_serializer
def integration_type_value_serializer(value):
    return value.value


# ---- Construct ----

USER_CONVERSIONS = AuditLogEntryDetailConversionGroup(
    CHANNEL_ID_CONVERSION,
    COUNT_CONVERSION,
    DELETE_MESSAGE_DURATION_CONVERSION,
    INTEGRATION_TYPE_CONVERSION,
    
    # Deprecations
    DELETE_MESSAGE_DURATION_DEPRECATED_DAYS_CONVERSION,
)



__all__ = ('CONVERSION_MESSAGE_REFERENCE_CONFIGURATION',)

from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion

from ..message_reference_configuration import MessageReferenceConfiguration


class CONVERSION_MESSAGE_REFERENCE_CONFIGURATION(Conversion):
    # Generic
    
    name = 'message_reference_configuration'
    name_aliases = None
    expected_types_messages = f'`{MessageReferenceConfiguration.__name__}`'
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    def set_merger(old_value, new_value):
        return old_value | new_value
    
    set_identifier = None
    set_type = None
    set_type_processor = None
    set_listing_identifier = None
    set_validator = None
    
    # Reading
    
    get_default = False
    get_processor = None
    
    
    # Serialization
    
    serializer_key = 'message_reference'
    
    
    def serializer_optional(value):
        if value:
            yield value.to_data()
    
    
    def serializer_required(value):
        return value.to_data()
    
    
    # Sorting
    
    sort_priority = -1



from .layouts import *

from .command import *
from .helpers import *
from .naming import *


__all__ = (
    *layouts.__all__,
    
    *command.__all__,
    *helpers.__all__,
    *naming.__all__,
)



__all__ = ()

from enum import Enum
from functools import partial as partial_func

from ...field_parsers import field_parser_factory, force_string_parser_factory, nullable_functional_parser_factory
from ...field_putters import (
    field_putter_factory, force_string_putter_factory, nullable_functional_optional_putter_factory
)
from ...localization.helpers import localized_dictionary_builder
from ...localization.utils import build_locale_dictionary, destroy_locale_dictionary

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN, VALUE_LENGTH_MAX, VALUE_LENGTH_MIN

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')


def validate_name(name):
    """
    Validates the given application command choice name.
    
    Parameters
    ----------
    name : `str`, `Enum`
        The name to validate.
    
    Returns
    -------
    name : `str`
        The validated name.
    
    Raises
    ------
    TypeError
        - If `name`'s type is incorrect.
    ValueError
        - If `name's length is out of the expected range.
    """
    if name is None:
        return ''
    
    if isinstance(name, Enum):
        processed_name = name.name
    else:
        processed_name = name
    
    # Check name type again in case enum name is not string, lol
    if not isinstance(processed_name, str):
        raise TypeError(
            f'`name` type can be `str`, `{Enum.__name__}`, '
            f'got {processed_name.__class__.__name__}; {processed_name!r}.'
        )
    
    name_length = len(processed_name)
    if (
        name_length and (
            (name_length < NAME_LENGTH_MIN) or
            (name_length > NAME_LENGTH_MAX)
        )
    ):
        raise ValueError(
            f'`name` length` can be in range '
            f'[{NAME_LENGTH_MIN}:{NAME_LENGTH_MAX}] '
            f'got {name_length!r}; {name!r}.'
        )
    
    return processed_name


# name_localizations

parse_name_localizations = nullable_functional_parser_factory(
    'name_localizations', build_locale_dictionary
)
put_name_localizations = nullable_functional_optional_putter_factory(
    'name_localizations', destroy_locale_dictionary
)
validate_name_localizations = partial_func(localized_dictionary_builder, parameter_name = 'name_localizations')



# value

parse_value = field_parser_factory('value')
put_value = field_putter_factory('value')


def validate_value(value):
    """
    Validates the given application command choice value.
    
    Parameters
    ----------
    value : `str`, `int`, `float`, `Enum`
        The value to validate.
    
    Returns
    -------
    value : `str`, `int`, `float`
        The validated value.
    
    Raises
    ------
    TypeError
        - If `value`'s type is incorrect.
    ValueError
        - If `value's length is out of the expected range.
    """
    if isinstance(value, Enum):
        processed_value = value.value
    else:
        processed_value = value
    
    if isinstance(processed_value, str):
        value_length = len(processed_value)
        if (
            (value_length < VALUE_LENGTH_MIN) or
            (value_length > VALUE_LENGTH_MAX)
        ):
            raise ValueError(
                f'`value` length` can be in range '
                f'[{VALUE_LENGTH_MIN}:{VALUE_LENGTH_MAX}]'
                f'got {value_length!r}; {value!r}.'
            )
    
    elif isinstance(processed_value, int):
        # No extra checks
        pass
    
    elif isinstance(processed_value, float):
        # No extra checks
        pass
    
    else:
        raise TypeError(
            f'`value` type can be `str`, `int`, `float`, `{Enum.__name__}`, '
            f'got {value.__class__.__name__}; {value!r}.'
        )
    
    return processed_value



__all__ = ()

from ...field_parsers import entity_id_parser_factory, preinstanced_parser_factory
from ...field_putters import entity_id_putter_factory, preinstanced_putter_factory
from ...field_validators import entity_id_validator_factory, preinstanced_validator_factory

from .preinstanced import ChannelType

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('channel_id')

# type

parse_type = preinstanced_parser_factory('type', ChannelType, ChannelType.unknown)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('channel_type', ChannelType)



__all__ = ()

from ...field_parsers import bool_parser_factory
from ...field_putters import bool_optional_putter_factory
from ...field_validators import bool_validator_factory

# emoji_pack_collectible

parse_emoji_pack_collectible = bool_parser_factory('is_emoji_pack_collectible', False)
put_emoji_pack_collectible = bool_optional_putter_factory('is_emoji_pack_collectible', False)
validate_emoji_pack_collectible = bool_validator_factory('emoji_pack_collectible', False)



__all__ = ('ParameterResult',)

from scarletio import RichAttributeErrorBaseType

from .result import COMMAND_RESULT_CODE_CONVERSION_FAILED, CommandResult


class ParameterResult(RichAttributeErrorBaseType):
    """
    Attributes
    ----------
    command_parameter : ``CommandParameter``
        The represented command parameter.
    fed_items : `None`, `list` of `tuple` (`str`, `object`) items
        Fed values to the parameter as results.
    """
    __slots__ = ('command_parameter', 'fed_items',)
    
    def __new__(cls, command_parameter):
        """
        Creates a new parameter result instance.
        
        Parameters
        ----------
        command_parameter : ``CommandParameter``
            The represented command parameter.
        """
        self = object.__new__(cls)
        self.command_parameter = command_parameter
        self.fed_items = None
        return self
    
    
    def is_satisfiable(self):
        """
        Returns whether the parameter can be satisfied.
        
        Returns
        -------
        is_satisfiable : `bool`
        """
        if self.fed_items is None:
            return True
        
        parameter = self.command_parameter.parameter
        if parameter.is_args() or parameter.is_kwargs():
            return True
        
        return False
    
    
    def is_satisfied(self):
        """
        Returns whether the parameter is satisfied.
        
        Returns
        -------
        is_satisfied : `bool`
        """
        if (self.fed_items is not None):
            return True
        
        parameter = self.command_parameter.parameter
        if parameter.has_default:
            return True
        
        if parameter.is_args() or parameter.is_kwargs():
            return True
        
        return False
    
    
    def feed(self, value):
        """
        Feeds a value with the given name.
        
        Parameters
        ----------
        value : `object`
            The value to feed.
        
        Returns
        ------
        command_result : `None`, ``CommandResult``
        """
        return self._feed(value, self.command_parameter.parameter.name)
    
    
    def feed_as(self, value, name):
        """
        Feeds a value with the given name.
        
        Parameters
        ----------
        value : `object`
            The value to feed.
        name : `str`
            The value's name.
        
        Returns
        ------
        command_result : `None`, ``CommandResult``
        """
        name = name.replace('-', '_')
        return self._feed(value, name)
    
    
    def _feed(self, value, name):
        """
        Feeds a value with the given name.
        
        This method is either called by ``.feed`` after auto filling `name`. Or by ``.feed_as`` after normalising
        the name.
        
        Parameters
        ----------
        value : `object`
            The value to feed.
        name : `str`
            The value's name.
        
        Returns
        ------
        command_result : `None`, ``CommandResult``
        """
        converter = self.command_parameter.get_converter()
        converted_value = converter(value)
        
        if (converted_value is None):
            return CommandResult(
                COMMAND_RESULT_CODE_CONVERSION_FAILED,
                self.command_parameter,
                value,
            )
        
        
        fed_items = self.fed_items
        if (fed_items is None):
            fed_items = []
            self.fed_items = fed_items
        
        fed_items.append((name, converted_value))
        return None
    
    
    def matches_name(self, name):
        """
        Returns whether the parameter matches the given name.
        
        Parameters
        ----------
        name : `str`
            The name to match.
        
        Returns
        -------
        is_matching : `bool`
        """
        command_parameter = self.command_parameter
        if command_parameter.display_name == name:
            return True
        
        if self.command_parameter.parameter.is_kwargs():
            return True
        
        return False
    
    
    def is_modifier(self):
        """
        Returns whether the parameter is a modifier parameter.
        
        Returns
        -------
        is_modifier : `str`
        """
        return self.command_parameter.is_modifier()
    
    
    
    def is_positional(self):
        """
        Returns whether the parameter is a positional parameter.
        
        Returns
        -------
        is_positional : `bool`
        """
        parameter = self.command_parameter.parameter
        if parameter.is_positional() or parameter.is_args():
            return True
        
        return False
    
    
    def has_fed_items(self):
        """
        Returns whether the parameter has fed values.
        
        Returns
        -------
        has_fed_items : `bool`
        """
        return (self.fed_items is not None)
    
    
    def get_fed_value(self):
        """
        Returns the first fed value.
        
        Returns
        -------
        value : `None`, `object`
        """
        fed_items = self.fed_items
        if (fed_items is not None):
            return fed_items[0][1]
    
    
    def get_fed_item(self):
        """
        Returns the first fed item.
        
        Returns
        -------
        item : `None`, `tuple` (`str`, `object`) item
        """
        fed_items = self.fed_items
        if (fed_items is not None):
            return fed_items[0]
    
    
    def iter_fed_values(self):
        """
        Iterates over the fed values of the parameter result.
        
        This method is an iterable generator.
        
        Yields
        ------
        value : `object`
        """
        fed_items = self.fed_items
        if (fed_items is not None):
            for fed_item in fed_items:
                yield fed_item[1]
    
    
    def iter_fed_items(self):
        """
        Iterates over the fed items of the parameter result.
        
        This method is an iterable generator.
        
        Yields
        ------
        item : `tuple` (`str`, `object`)
        """
        fed_items = self.fed_items
        if (fed_items is not None):
            yield from fed_items
    
    
    def put_into(self, positional_parameters, keyword_parameters):
        """
        Puts the value of the parameter into the given containers representing a function's call parameters.
        
        Parameters
        ----------
        positional_parameters : `list` of `object`
            Positional parameters to call the respective function with.
        keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters to call the respective function with.
        
        Returns
        -------
        was_put : `bool`
            Whether putting values into the parameters was successful.
        """
        parameter = self.command_parameter.parameter
        if parameter.is_positional():
            if self.has_fed_items():
                positional_parameters.append(self.get_fed_value())
                return True
            
            if parameter.has_default:
                positional_parameters.append(parameter.default)
                return True
            
            return False
        
        
        if parameter.is_keyword_only():
            if self.has_fed_items():
                key, value = self.get_fed_item()
                keyword_parameters[key] = value
                return True
            
            if parameter.has_default:
                keyword_parameters[parameter.name] = parameter.default
                return True
            
            return False
        
        
        if parameter.is_args():
            positional_parameters.extend(self.iter_fed_values())
            return True
        
        
        if parameter.is_kwargs():
            keyword_parameters.update(self.iter_fed_items())
            return True
        
        
        return False



__all__ = ('InteractionMetadataApplicationCommand',)

from warnings import warn

from scarletio import copy_docs

from ...application_command import ApplicationCommandTargetType

from .application_command_autocomplete import InteractionMetadataApplicationCommandAutocomplete
from .fields import (
    parse_resolved, parse_target_id, parse_target_type, put_resolved, put_target_id, put_target_type,
    validate_resolved, validate_target_id, validate_target_type
)


class InteractionMetadataApplicationCommand(InteractionMetadataApplicationCommandAutocomplete):
    """
    Interaction metadata used when the interaction was triggered by an application command.
    
    Parameters
    ----------
    id : `int`
        The represented application command's identifier number.
    
    name : `str`
        The represented application command's name.
    
    options : `None`, `tuple` of ``InteractionOption``
        Application command option representations. Like sub-command or parameter.
    
    resolved : `None`, ``Resolved``
        Contains the received entities.
    
    target_id : `int`
        The interaction's target's identifier. Applicable for context commands.
    
    target_type : ``ApplicationCommandTargetType``
        The invoked application command's target type.
    """
    __slots__ = ('resolved', 'target_id', 'target_type')
    
    
    def __new__(
        cls,
        *,
        application_command_id = ...,
        name = ...,
        options = ...,
        resolved = ...,
        target_id = ...,
        target_type = ...,
    ):
        """
        Creates a new interaction metadata from the given parameters.
        
        Parameters
        ----------
        application_command_id : `int`, Optional (Keyword only)
            The represented application command's identifier number.
        
        name : `str`, Optional (Keyword only)
            The represented application command's name.
        
        options : `None`, `tuple` of ``InteractionOption``, Optional (Keyword only)
            Application command option representations. Like sub-command or parameter.
        
        resolved : `None`, ``Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        target_id : `int`, Optional (Keyword only)
            The interaction's target's identifier. Applicable for context commands.
        
        target_type : `ApplicationCommandTargetType | None | int`, Optional (Keyword only)
            The invoked application command's target type.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # resolved
        if resolved is ...:
            resolved = None
        else:
            resolved = validate_resolved(resolved)
        
        # target_id
        if target_id is ...:
            target_id = 0
        else:
            target_id = validate_target_id(target_id)
        
        # target_type
        if target_type is ...:
            target_type = ApplicationCommandTargetType.none
        else:
            target_type = validate_target_type(target_type)
        
        # Construct
        self = InteractionMetadataApplicationCommandAutocomplete.__new__(
            cls,
            application_command_id = application_command_id,
            name = name,
            options = options,
        )
        self.resolved = resolved
        self.target_id = target_id
        self.target_type = target_type
        return self
    
    
    @classmethod
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete._create_empty)
    def _create_empty(cls):
        self = super(InteractionMetadataApplicationCommand, cls)._create_empty()
        self.resolved = None
        self.target_id = 0
        self.target_type = ApplicationCommandTargetType.none
        return self
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete.copy)
    def copy(self):
        new = InteractionMetadataApplicationCommandAutocomplete.copy(self)
        
        resolved = self.resolved
        if (resolved is not None):
            resolved = resolved.copy()
        new.resolved = resolved
        
        new.target_id = self.target_id 
        new.target_type = self.target_type
        
        return new
    
    
    def copy_with(
        self,
        *,
        application_command_id = ...,
        id = ...,
        name = ...,
        options = ...,
        resolved = ...,
        target_id = ...,
        target_type = ...,
    ):
        """
        Copies the interaction metadata with the given fields.
        
        Parameters
        ----------
        application_command_id : `int`, Optional (Keyword only)
            The represented application command's identifier number.
        
        name : `str`, Optional (Keyword only)
            The represented application command's name.
        
        options : `None`, `tuple` of ``InteractionOption``, Optional (Keyword only)
            Application command option representations. Like sub-command or parameter.
        
        resolved : `None`, ``Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        target_id : `int`, Optional (Keyword only)
            The interaction's target's identifier. Applicable for context commands.
        
        target_type : `ApplicationCommandTargetType | None | int`, Optional (Keyword only)
            The invoked application command's target type.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # id - Deprecated
        if id is not ...:
            warn(
                (
                    f'`{type(self).__name__}.copy_with`\' `id` parameter is deprecated '
                    f'and will be removed in 2024 December. '
                    f'Please use `application_command_id` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            application_command_id = id
        
        # resolved
        if resolved is ...:
            resolved = self.resolved
            if (resolved is not None):
                resolved = resolved.copy()
        else:
            resolved = validate_resolved(resolved)
        
        # target_id
        if target_id is ...:
            target_id = self.target_id
        else:
            target_id = validate_target_id(target_id)
        
        # target_type
        if target_type is ...:
            target_type = self.target_type
        else:
            target_type = validate_target_type(target_type)
        
        # Construct
        new = InteractionMetadataApplicationCommandAutocomplete.copy_with(
            self,
            application_command_id = application_command_id,
            name = name,
            options = options,
        )
        new.resolved = resolved
        new.target_id = target_id
        new.target_type = target_type
        return new
    
    
    @classmethod
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete.from_data)
    def from_data(cls, data, guild_id = 0):
        self = super(InteractionMetadataApplicationCommand, cls).from_data(data, guild_id)
        self.resolved = parse_resolved(data, guild_id)
        self.target_id = parse_target_id(data)
        self.target_type = parse_target_type(data)
        return self
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete.to_data)
    def to_data(self, *, defaults = False, guild_id = 0):
        data = InteractionMetadataApplicationCommandAutocomplete.to_data(
            self, defaults = defaults, guild_id = guild_id
        )
        put_resolved(self.resolved, data, defaults, guild_id = guild_id)
        put_target_id(self.target_id, data, defaults)
        put_target_type(self.target_type, data, defaults)
        return data
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete._put_attribute_representations_into)
    def _put_attribute_representations_into(self, repr_parts):
        field_added = InteractionMetadataApplicationCommandAutocomplete._put_attribute_representations_into(
            self, repr_parts
        )
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' resolved = ')
            repr_parts.append(repr(resolved))
        
        # target_id
        target_id = self.target_id
        if target_id:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' target_id = ')
            repr_parts.append(repr(target_id))
        
        # target_type
        target_type = self.target_type
        if target_type:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' target_type = ')
            repr_parts.append(target_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(target_type.value))
        
        return field_added
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete.__hash__)
    def __hash__(self):
        hash_value = InteractionMetadataApplicationCommandAutocomplete.__hash__(self)
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            hash_value ^= hash(resolved)
        
        # target_id
        hash_value ^= self.target_id
        
        # target_type
        hash_value ^= hash(self.target_type) << 4
        
        return hash_value
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not InteractionMetadataApplicationCommandAutocomplete._is_equal_same_type(self, other):
            return False
        
        # resolved
        if self.resolved != other.resolved:
            return False
        
        # target_id
        if self.target_id != other.target_id:
            return False
        
        # target_type
        if self.target_type is not other.target_type:
            return False
        
        return True



from .constants import *
from .emoji_counts import *
from .fields import *
from .flags import *
from .guild import *
from .guild_premium_perks import *
from .helpers import *
from .preinstanced import *
from .sticker_counts import *
from .utils import *


__all__ = (
    *constants.__all__,
    *emoji_counts.__all__,
    *fields.__all__,
    *flags.__all__,
    *guild.__all__,
    *guild_premium_perks.__all__,
    *helpers.__all__,
    *preinstanced.__all__,
    *sticker_counts.__all__,
    *utils.__all__,
)


from ....utils.module_deprecation import deprecated_import
deprecated_import(ExplicitContentFilterLevel, 'ContentFilterLevel')
deprecated_import(MfaLevel, 'MFA')



__all__ = ('PreinstancedBase',)

from warnings import warn

from scarletio import RichAttributeErrorBaseType

from .preinstanced_meta import PreinstancedMeta


class PreinstancedBase(RichAttributeErrorBaseType, metaclass = PreinstancedMeta, base_type = True):
    """
    Base type for other preinstanced types.
    
    Attributes
    ----------
    name : `str`
        The name of the instance.
    
    value : `int | str`
        The value representing the instance.
    """
    __slots__ = ('name', 'value')
    
    
    def __new__(cls, value, name = None):
        """
        Creates a new preinstanced instance.
        
        Parameters
        ----------
        value : ``.VALUE_TYPE``
            The value of the preinstanced object.
        
        name : `None | str` = `None`, Optional
            The object's name.
        """
        if name is None:
            if cls.VALUE_TYPE is str:
                name = value
            else:
                name = cls.NAME_DEFAULT
        
        self = object.__new__(cls)
        self.value = value
        self.name = name
        return self
    
    
    @classmethod
    def get(cls, value):
        """
        Returns the value's representation. If the value is already preinstanced, returns that, else creates a new
        object.
        
        Parameters
        ----------
        value : `None`, ``.VALUE_TYPE``
            The value to get it's representation.
        
        Returns
        -------
        self : `type<cls>`
        """
        warn(
            (
                f'`{cls.__name__}` is deprecated and will be removed in 2025 September.'
                f'Please use `{cls.__name__}(value)` instead.',
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return cls(value)
    
    
    def __lt__(self, other):
        """Returns self < other."""
        self_type = type(self)
        other_type = type(other)
        
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        return (self.value < other_value)
    
    
    def __gt__(self, other):
        """Returns self > other."""
        self_type = type(self)
        other_type = type(other)
        
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        return (self.value > other_value)
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if self is other:
            return True
        
        self_type = type(self)
        other_type = type(other)
        
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        return (self.value == other_value)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        return hash(self.value)
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # value
        value = self.value
        repr_parts.append(' value = ')
        repr_parts.append(repr(value))
        
        # name | skip it, if same as value; some sub-types want to drop `.name` from `repr()` if its same as `.value`.
        name = self.name
        if (self.VALUE_TYPE is not str) or (value != name):
            repr_parts.append(', name = ')
            repr_parts.append(repr(self.name))
        
        # extra as available
        self._put_repr_parts_into(repr_parts)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_repr_parts_into(self, repr_parts):
        """
        Appends the representation parts.
        
        Parameters
        ----------
        repr_parts : `list<str>`
            Representation parts to extend.
        """
        return



__all__ = ('Attachment', )

from ...bases import DiscordEntity
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .constants import DURATION_DEFAULT
from .fields import (
    parse_application, parse_clip_created_at, parse_clip_users, parse_content_type, parse_description, parse_duration,
    parse_flags, parse_height, parse_id, parse_name, parse_proxy_url, parse_size, parse_temporary, parse_title,
    parse_url, parse_waveform, parse_width, put_application, put_clip_created_at, put_clip_users,
    put_content_type, put_description, put_duration, put_flags, put_height, put_id,
    put_name, put_proxy_url, put_size, put_temporary, put_title, put_url,
    put_waveform, put_width, validate_application, validate_clip_created_at, validate_clip_users,
    validate_content_type, validate_description, validate_duration, validate_flags, validate_height, validate_id,
    validate_name, validate_proxy_url, validate_size, validate_temporary, validate_title, validate_url,
    validate_waveform, validate_width
)
from .flags import AttachmentFlag


PRECREATE_FIELDS = {
    'application': ('application', validate_application),
    'clip_created_at': ('clip_created_at', validate_clip_created_at),
    'clip_users': ('clip_users', validate_clip_users),
    'content_type': ('content_type', validate_content_type),
    'description': ('description', validate_description),
    'duration': ('duration', validate_duration),
    'flags': ('flags', validate_flags),
    'height': ('height', validate_height),
    'name': ('name', validate_name),
    'proxy_url': ('proxy_url', validate_proxy_url),
    'size': ('size', validate_size),
    'temporary': ('temporary', validate_temporary),
    'title': ('title', validate_title),
    'url': ('url', validate_url),
    'waveform': ('waveform', validate_waveform),
    'width': ('width', validate_width),
}


class Attachment(DiscordEntity):
    """
    Represents an attachment of a ``Message``.
    
    Attributes
    ----------
    application : `None | Application`
        The application in the attachment if recognized. (Only for clips for now.)
    
    clip_created_at : `None | DateTime`
        When the clip was created. Applicable if the attachment is a clip.
    
    clip_users : `None | tuple<ClientUserBase>`
        The users in the clip. Applicable if the attachment is a clip.
    
    content_type : `None`, `str`
        The attachment's media type.
    
    description : `None`, `str`
        Description for the file.
        
        > Max 1024 characters.
    
    duration : `float`
        The attachment's duration in seconds. Applicable for voice messages only.
        
        > Defaults to `0.0`.
    
    flags : ``AttachmentFlag``
        Flags of the attachment.
    
    height : `int`
        The height of the attachment if applicable.
        
        > Defaults to `0`.
    
    id : `int`
        The unique identifier number of the attachment.
    
    name : `str`
        The name of the attachment.
        Special and unicode characters are excluded from an attachment's name.
    
    proxy_url : `str`
        Proxied url of the attachment.
    
    size : `int`
        The attachment's size in bytes.
    
    temporary : `bool`
        Whether the attachment is temporary and is removed after a set period of time.
        
        Temporary attachments are guaranteed to be available as long as their message itself exists.
        
        > Defaults to `False`.
    
    title : `None`, `str`
        The attachment's title.
        Present if any characters were excluded from ``.name``.
    
    url : `str`
        The attachment's url.
    
    waveform : `None`, `str`
        Base64 encoded bytearray representing a sampled waveform. Applicable for voice messages only.
        
        > Defaults to `None`.
    
    width : `int`
        The attachment's width if applicable.
        
        > Defaults to `0`.
    """
    __slots__ = (
        'application', 'clip_created_at', 'clip_users', 'content_type', 'description', 'duration', 'flags', 'height',
        'name', 'proxy_url', 'size', 'temporary', 'title', 'url', 'waveform', 'width'
    )
    
    def __new__(
        cls,
        *,
        application = ...,
        clip_created_at = ...,
        clip_users = ...,
        content_type = ...,
        description = ...,
        duration = ...,
        flags = ...,
        height = ...,
        name = ...,
        size = ...,
        temporary = ...,
        title = ...,
        url = ...,
        waveform = ...,
        width = ...,
    ):
        """
        Creates a new partial attachment.
        
        Parameters
        ----------
        application : `None | Application`, Optional (Keyword only)
            The application in the attachment if recognized.
        
        clip_created_at : `None | DateTime`, Optional (Keyword only)
            When the clip was created. Applicable if the attachment is a clip.
        
        clip_users : `None | iterable<ClientUserBase>`, Optional (Keyword only)
            The users in the clip. Applicable if the attachment is a clip.
        
        content_type : `None`, `str`, Optional (Keyword only)
            The attachment's media type.
        
        description : `None`, `str`, Optional (Keyword only)
            Description for the file.
        
        duration : `float`, Optional (Keyword only)
            The attachment's duration in seconds.
        
        flags : ``AttachmentFlag``, `int`, Optional (Keyword only)
            The attachment's flags.
        
        height : `int`, Optional (Keyword only)
            The height of the attachment if applicable.
        
        name : `str`, Optional (Keyword only)
            The name of the attachment.
        
        size : `int`, Optional (Keyword only)
            The attachment's size in bytes.
        
        temporary : `bool`, Optional (Keyword only)
            Whether the attachment is temporary and is removed after a set period of time.
        
        title : `None`, `str`, Optional (Keyword only)
            The attachment's title.
        
        url : `str`, Optional (Keyword only)
            The attachment's url.
        
        waveform : `None`, `str`, Optional (Keyword only)
            Base64 encoded bytearray representing a sampled waveform.
        
        width : `int`, Optional (Keyword only)
            The attachment's width if applicable.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application
        if application is ...:
            application = None
        else:
            application = validate_application(application)
        
        # clip_created_at
        if clip_created_at is ...:
            clip_created_at = None
        else:
            clip_created_at = validate_clip_created_at(clip_created_at)
        
        # clip_users
        if clip_users is ...:
            clip_users = None
        else:
            clip_users = validate_clip_users(clip_users)
        
        # content_type
        if content_type is ...:
            content_type = None
        else:
            content_type = validate_content_type(content_type)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # duration
        if duration is ...:
            duration = DURATION_DEFAULT
        else:
            duration = validate_duration(duration)
        
        # flags
        if flags is ...:
            flags = AttachmentFlag()
        else:
            flags = validate_flags(flags)
        
        # height
        if height is ...:
            height = 0
        else:
            height = validate_height(height)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # size
        if size is ...:
            size = 0
        else:
            size = validate_size(size)
        
        # temporary
        if temporary is ...:
            temporary = False
        else:
            temporary = validate_temporary(temporary)
        
        # title
        if title is ...:
            title = None
        else:
            title = validate_title(title)
        
        # url
        if url is ...:
            url = ''
        else:
            url = validate_url(url)
        
        # waveform
        if waveform is ...:
            waveform = None
        else:
            waveform = validate_waveform(waveform)
        
        # width
        if width is ...:
            width = 0
        else:
            width = validate_width(width)
        
        # Construct
        
        self = object.__new__(cls)
        self.application = application
        self.clip_created_at = clip_created_at
        self.clip_users = clip_users
        self.content_type = content_type
        self.description = description
        self.duration = duration
        self.flags = flags
        self.height = height
        self.id = 0
        self.name = name
        self.proxy_url = None
        self.size = size
        self.temporary = temporary
        self.title = title
        self.url = url
        self.waveform = waveform
        self.width = width
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an attachment object from the attachment data included inside of a ``Message`'s.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received attachment data.
        """
        self = object.__new__(cls)
        self.application = parse_application(data)
        self.clip_created_at = parse_clip_created_at(data)
        self.clip_users = parse_clip_users(data)
        self.content_type = parse_content_type(data)
        self.description = parse_description(data)
        self.duration = parse_duration(data)
        self.flags = parse_flags(data)
        self.height = parse_height(data)
        self.id = parse_id(data)
        self.name = parse_name(data)
        self.proxy_url = parse_proxy_url(data)
        self.size = parse_size(data)
        self.temporary = parse_temporary(data)
        self.title = parse_title(data)
        self.url = parse_url(data)
        self.waveform = parse_waveform(data)
        self.width = parse_width(data)
        return self
    
    
    def __repr__(self):
        """Returns the representation of the attachment."""
        repr_parts = ['<', type(self).__name__,]
        
        attachment_id = self.id
        if attachment_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(attachment_id))
            
            if self.temporary:
                repr_parts.append(' (temporary)')
            
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # If has title
        title = self.title
        if (title is not None):
            repr_parts.append(', title = ')
            repr_parts.append(repr(title))
        
        # Extra if audio
        duration = self.duration
        if duration:
            repr_parts.append(', duration = ')
            repr_parts.append(format(duration, '.02f'))
        
        # Extra if image
        width = self.width
        height = self.height
        if width and height:
            repr_parts.append(', size = ')
            repr_parts.append(repr(width))
            repr_parts.append('x')
            repr_parts.append(repr(height))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two attachments are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # shortcut
        if self is other:
            return True
        
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            # Only compare `.id` and `.proxy_url` if both attachment is not partial.
            
            # id
            if self.id != other.id:
                return False
            
            # proxy_url
            if self.proxy_url != other.proxy_url:
                return False
        
        # application
        if self.application != other.application:
            return False
        
        # clip_created_at
        if self.clip_created_at != other.clip_created_at:
            return False
        
        # clip_users
        if self.clip_users != other.clip_users:
            return False
        
        # content_type
        if self.content_type != other.content_type:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # duration
        if self.duration != other.duration:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # height
        if self.height != other.height:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # size
        if self.size != other.size:
            return False
        
        # temporary
        if self.temporary != other.temporary:
            return False
        
        # title
        if self.title != other.title:
            return False
        
        # url
        if self.url != other.url:
            return False
        
        # waveform
        if self.waveform != other.waveform:
            return False
        
        # width
        if self.width != other.width:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the attachment."""
        hash_value = 0
        
        # application
        application = self.application
        if (application is not None):
            hash_value ^= hash(application)
        
        # clip_created_at
        clip_created_at = self.clip_created_at
        if (clip_created_at is not None):
            hash_value ^= hash(clip_created_at)
        
        # clip_users
        clip_users = self.clip_users
        if (clip_users is not None):
            hash_value ^= len(clip_users) << 16
            for clip_user in clip_users:
                hash_value ^= hash(clip_user)
        
        # content_type
        content_type = self.content_type
        if (content_type is not None):
            hash_value ^= hash(content_type)
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # duration
        duration = self.duration
        if duration:
            hash_value ^= hash(duration)
        
        # flags
        hash_value ^= self.flags << 3
        
        # height
        hash_value ^= self.height
        
        # id
        attachment_id = self.id
        if attachment_id:
            hash_value ^= attachment_id
            
            # proxy_url
            # Add proxy url to hash only if we have attachment id.
            proxy_url = self.proxy_url
            if (proxy_url is not None):
                hash_value ^= hash(proxy_url)
        
        # name
        name = self.name
        if (description is None) or (name != description):
            hash_value ^= hash(name)
        
        # size
        hash_value ^= self.size << 17
        
        # temporary
        hash_value ^= self.temporary << 16
        
        # title
        title = self.title
        if (title is not None):
            hash_value ^= hash(title)
        
        # url
        hash_value ^= hash(self.url)
        
        # waveform
        waveform = self.waveform
        if (waveform is not None):
            hash_value ^= hash(waveform)
        
        # width
        hash_value ^= self.width << 8
        
        return hash_value
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Tries to convert the attachment back to json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_application(self.application, data, defaults)
        put_clip_created_at(self.clip_created_at, data, defaults)
        put_clip_users(self.clip_users, data, defaults)
        put_content_type(self.content_type, data, defaults)
        put_description(self.description, data, defaults)
        put_duration(self.duration, data, defaults)
        put_flags(self.flags, data, defaults)
        put_height(self.height, data, defaults)
        put_name(self.name, data, defaults)
        put_size(self.size, data, defaults)
        put_temporary(self.temporary, data, defaults)
        put_title(self.title, data, defaults)
        put_url(self.url, data, defaults)
        put_waveform(self.waveform, data, defaults)
        put_width(self.width, data, defaults)
        
        if include_internals:
            put_id(self.id, data, defaults)
            put_proxy_url(self.proxy_url, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Returns a copy of the attachment.
        
        > The copy will not include the internal fields of it.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.application = self.application
        new.clip_created_at = self.clip_created_at
        
        clip_users = self.clip_users
        if (clip_users is not None):
            clip_users = (*clip_users,)
        new.clip_users = clip_users
        
        new.content_type = self.content_type
        new.description = self.description
        new.duration = self.duration
        new.flags = self.flags
        new.height = self.height
        new.id = 0
        new.name = self.name
        new.proxy_url = None
        new.size = self.size
        new.temporary = self.temporary
        new.title = self.title
        new.url = self.url
        new.waveform = self.waveform
        new.width = self.width
        return new
    
    
    def copy_with(
        self,
        *,
        application = ...,
        clip_created_at = ...,
        clip_users = ...,
        content_type = ...,
        description = ...,
        duration = ...,
        flags = ...,
        height = ...,
        name = ...,
        size = ...,
        temporary = ...,
        title = ...,
        url = ...,
        waveform = ...,
        width = ...,
    ):
        """
        Returns a copy of the attachment. The attributes of the copy are modified based on the given values.
        
        > The copy will not include the internal fields of it.
        
        Parameters
        ----------
        application : `None | Application`, Optional (Keyword only)
            The application in the attachment if recognized.
        
        clip_created_at : `None | DateTime`, Optional (Keyword only)
            When the clip was created. Applicable if the attachment is a clip.
        
        clip_users : `None | iterable<ClientUserBase>`, Optional (Keyword only)
            The users in the clip. Applicable if the attachment is a clip.
        
        content_type : `None`, `str`, Optional (Keyword only)
            The attachment's media type.
        
        description : `None`, `str`, Optional (Keyword only)
            Description for the file.
        
        duration : `float`, Optional (Keyword only)
            The attachment's duration in seconds.
        
        flags : ``AttachmentFlag``, `int`, Optional (Keyword only)
            The attachment's flags.
        
        height : `int`, Optional (Keyword only)
            The height of the attachment if applicable.
        
        name : `str`, Optional (Keyword only)
            The name of the attachment.
        
        size : `int`, Optional (Keyword only)
            The attachment's size in bytes.
        
        temporary : `bool`, Optional (Keyword only)
            Whether the attachment is temporary and is removed after a set period of time.
        
        title : `None`, `str`, Optional (Keyword only)
            The attachment's title.
        
        url : `str`, Optional (Keyword only)
            The attachment's url.
        
        waveform : `None`, `str`, Optional (Keyword only)
            Base64 encoded bytearray representing a sampled waveform.
        
        width : `int`, Optional (Keyword only)
            The attachment's width if applicable.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        # application
        if application is ...:
            application = self.application
        else:
            application = validate_application(application)
        
        # clip_created_at
        if clip_created_at is ...:
            clip_created_at = self.clip_created_at
        else:
            clip_created_at = validate_clip_created_at(clip_created_at)
        
        # clip_users
        if clip_users is ...:
            clip_users = self.clip_users
            if (clip_users is not None):
                clip_users = (*clip_users,)
        else:
            clip_users = validate_clip_users(clip_users)
        
        # content_type
        if content_type is ...:
            content_type = self.content_type
        else:
            content_type = validate_content_type(content_type)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # duration
        if duration is ...:
            duration = self.duration
        else:
            duration = validate_duration(duration)
        
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # height
        if height is ...:
            height = self.height
        else:
            height = validate_height(height)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # size
        if size is ...:
            size = self.size
        else:
            size = validate_size(size)
        
        # temporary
        if temporary is ...:
            temporary = self.temporary
        else:
            temporary = validate_temporary(temporary)
        
        # title
        if title is ...:
            title = self.title
        else:
            title = validate_title(title)
        
        # url
        if url is ...:
            url = self.url
        else:
            url = validate_url(url)
        
        # waveform
        if waveform is ...:
            waveform = self.waveform
        else:
            waveform = validate_waveform(waveform)
        
        # width
        if width is ...:
            width = self.width
        else:
            width = validate_width(width)
        
        # Construct
        
        new = object.__new__(type(self))
        new.application = application
        new.clip_created_at = clip_created_at
        new.clip_users = clip_users
        new.content_type = content_type
        new.description = description
        new.duration = duration
        new.flags = flags
        new.height = height
        new.id = 0
        new.name = name
        new.proxy_url = None
        new.size = size
        new.temporary = temporary
        new.title = title
        new.url = url
        new.waveform = waveform
        new.width = width
        return new
    
    
    @classmethod
    def precreate(
        cls,
        attachment_id,
        **keyword_parameters,
    ):
        """
        Precreates an attachment. Since attachments are not cached, this method just a ``.__new__`` alternative.
        
        Parameters
        ----------
        attachment_id : `int`
            The attachment's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional parameters defining how the attachment's fields should be set.
        
        Other Parameters
        ----------------
        application : `None | Application`, Optional (Keyword only)
            The application in the attachment if recognized.
        
        clip_created_at : `None | DateTime`, Optional (Keyword only)
            When the clip was created. Applicable if the attachment is a clip.
        
        clip_users : `None | iterable<ClientUserBase>`, Optional (Keyword only)
            The users in the clip. Applicable if the attachment is a clip.
        
        content_type : `None`, `str`, Optional (Keyword only)
            The attachment's media type.
        
        description : `None`, `str`, Optional (Keyword only)
            Description for the file.
        
        duration : `float`, Optional (Keyword only)
            The attachment's duration in seconds.
        
        flags : ``AttachmentFlag``, `int`, Optional (Keyword only)
            The attachment's flags.
        
        height : `int`, Optional (Keyword only)
            The height of the attachment if applicable.
        
        name : `str`, Optional (Keyword only)
            The name of the attachment.
        
        size : `int`, Optional (Keyword only)
            The attachment's size in bytes.
        
        proxy_url : `None`, `str`, Optional (Keyword only)
            Proxied url of the attachment.
        
        temporary : `bool`, Optional (Keyword only)
            Whether the attachment is temporary and is removed after a set period of time.
        
        title : `None`, `str`, Optional (Keyword only)
            The attachment's title.
            
        url : `str`, Optional (Keyword only)
            The attachment's url.
        
        waveform : `None`, `str`, Optional (Keyword only)
            Base64 encoded bytearray representing a sampled waveform.
        
        width : `int`, Optional (Keyword only)
            The attachment's width if applicable.
        
        Returns
        -------
        self : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        attachment_id = validate_id(attachment_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        
        else:
            processed = None
        
        # Construct
        
        self = object.__new__(cls)
        self.application = None
        self.clip_created_at = None
        self.clip_users = None
        self.content_type = None
        self.description = None
        self.duration = DURATION_DEFAULT
        self.flags = AttachmentFlag()
        self.height = 0
        self.id = attachment_id
        self.name = ''
        self.proxy_url = None
        self.size = 0
        self.temporary = False
        self.title = None
        self.url = ''
        self.waveform = None
        self.width = 0
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @property
    def display_name(self):
        """
        Returns the displayed name of the attachment.
        
        Displayed name is not the upload name, but something hata tries to construct to imitate it as close as possible.
        `attachment.name` do not include unicodes and `attachment.title` that was added to address this issue is
        not including the extension. So if there are any unicodes in extension they are omitted.
        To make this worse, uploaded attachment through interaction commands do not even have `.title`.
        
        Include a lot of cursing in this section.
        
        Returns
        -------
        display_name : `str`
        """
        title = self.title
        name = self.name
        
        if (title is None):
            return name
        
        extension_dot_index = name.find('.')
        if extension_dot_index == -1:
            return title
        
        return title + name[extension_dot_index:]
    
    
    @property
    def content_created_at(self):
        """
        Returns when the attachment's content was created if known.
        
        Returns
        -------
        modified_at : `DateTime`
        """
        content_created_at = self.clip_created_at
        if (content_created_at is not None):
            return content_created_at
        
        return self.created_at
    
    
    def iter_clip_users(self):
        """
        Iterates over the users in the clip. Applicable if the attachment is a clip.
        
        This method is an iterable generator.
        
        Yields
        ------
        clip_user : ``ClientUserBase``
        """
        clip_users = self.clip_users
        if (clip_users is not None):
            yield from clip_users



__all__ = ()

from ...field_parsers import nullable_date_time_parser_factory
from ...field_putters import nullable_date_time_optional_putter_factory
from ...field_validators import duration_validator_factory, nullable_date_time_validator_factory

# direct_message_spam_detected_at

parse_direct_message_spam_detected_at = nullable_date_time_parser_factory('dm_spam_detected_at')
put_direct_message_spam_detected_at = nullable_date_time_optional_putter_factory('dm_spam_detected_at')
validate_direct_message_spam_detected_at = nullable_date_time_validator_factory('direct_message_spam_detected_at')

# direct_messages_disabled_duration

validate_direct_messages_disabled_duration = duration_validator_factory('direct_messages_disabled_duration')

# direct_messages_disabled_until

parse_direct_messages_disabled_until = nullable_date_time_parser_factory('dms_disabled_until')
put_direct_messages_disabled_until = nullable_date_time_optional_putter_factory('dms_disabled_until')
validate_direct_messages_disabled_until = nullable_date_time_validator_factory('direct_messages_disabled_until')

# invites_disabled_duration

validate_invites_disabled_duration = duration_validator_factory('invites_disabled_duration')

# invites_disabled_until

parse_invites_disabled_until = nullable_date_time_parser_factory('invites_disabled_until')
put_invites_disabled_until = nullable_date_time_optional_putter_factory('invites_disabled_until')
validate_invites_disabled_until = nullable_date_time_validator_factory('invites_disabled_until')

# raid_detected_at

parse_raid_detected_at = nullable_date_time_parser_factory('raid_detected_at')
put_raid_detected_at = nullable_date_time_optional_putter_factory('raid_detected_at')
validate_raid_detected_at = nullable_date_time_validator_factory('raid_detected_at')
